/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-agent.h,v 1.61 2002/06/22 01:47:57 actmodern Exp $
 *
 * Copyright 2002 Thamer Alharbash
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

/* XXX -- ok this is getting ridiculously large
 * we need to split these into smaller header
 * files. */

#ifndef DHCP_AGENT_H
# define DHCP_AGENT_H

#define __STDC_FORMAT_MACROS /* we need the stdint type macros. */

# include <stdlib.h>
# include <stdio.h>

# include <sys/types.h>
# include <sys/time.h>
# include <sys/socket.h>
# include <sys/stat.h>

# include <netinet/in.h>
# include <arpa/inet.h>

# include <netdb.h>
# include <string.h>
# include <errno.h>
# include <fcntl.h>
# include <ctype.h>
# include <unistd.h>
# include <syslog.h>
# include <signal.h>
# include <time.h>
# include <dnet.h>
# include <pcap.h>

# include <config.h>

/* must be placed in main source of each binary.
 * if not available. */
extern const char *__progname;

#if defined(HAVE_GETOPT_H)

# include <getopt.h>

#endif /* HAVE_GETOPT_H */

#ifdef HAVE_PRIMACROS_H

#include <primacros.h>

#endif /* HAVE_PRIMACROS_H */

#if defined(HAVE_STDARG_H)

# include <stdarg.h>

# else

# if defined(HAVE_SYS_VARARGS_H)

#  include <varargs.h>

# endif /* HAVE_VARARGS_H */

#endif /* HAVE_STDARG_H */

# if defined(HAVE_SYS_UTSNAME_H)

# include <sys/utsname.h>

# endif /* HAVE_UTSNAME_H */


/* * * * * * * * * * *
 * Limit definitions *
 * * * * * * * * * * */

/* Limit on size of strings we'd be willing to handle. */

# define MAX_STRING_LEN 256

/* Generic size for buffers we use.
 * XXX -- fix generic buffers into sensible sizes.
 */

# define GENERIC_BUFFSIZE 256

/* log string buffer */

# define MSG_BUFFER_SIZE 256

/* Snaplen. Use the largest size since DHCP packets could
 * get pretty big with all the additional options.
 * 
 * We'll assume the MTU is 1500 for now. (FIXME)
 * 
 */

# define DEFAULT_MTU 1500
# define SNAPLEN DEFAULT_MTU /* XXX -- should be called DEFAULT_SNAPLEN.
                              * rawnet makes an assumption on incoming
                              * packets this is broken. fixme.
                              * 
                              */

/* max options handled. */

# define MAX_OPTIONS_HANDLED 62

/* sensible default for max descriptors -- we use this if we
   can't find it through other means (sysconf/getrusage) */

#define SENSIBLE_DESCRIPTOR_MAX 128

/* type string limits. */

# define UINT32_T_STRING_SIZE 11
# define INT32_T_STRING_SIZE UINT32_T_STRING_SIZE /* the same. */
# define UINT16_T_STRING_SIZE 6
# define INT16_T_STRING_SIZE UINT16_T_STRING_SIZE
# define BYTE_STRING_SIZE 5

/* Reasonable defaults in case services db isn't up to date. */

#define BOOTP_CLIENT 68
#define BOOTP_SERVER 67

/* Solaris doesn't have INADDR_NONE */

#ifndef INADDR_NONE
#define INADDR_NONE (uint32_t) -1
#endif

/* Verbosity levels. */

#define QUIET_VERBOSITY_LEVEL 0
#define ERROR_VERBOSITY_LEVEL 1
#define NORMAL_VERBOSITY_LEVEL 2
#define WARNING_VERBOSITY_LEVEL 3
#define DEBUG_VERBOSITY_LEVEL 4
#define MAX_VERBOSITY_LEVEL DEBUG_VERBOSITY_LEVEL

/* default configuration */

#define DISCOVER_OFFER_RETRIES 3

/* * * * * * * * * * *
 * Data structures.  *
 * * * * * * * * * * */

/* List data structure *
 * for linked lists.   */

typedef struct list {
    struct list *next;
    void *data;
} list_t;

/* DHCP fixed header -- before dhcp options. */

/* Size of header independant of any padding
 * sizeof() the structure may return. */

#define DHCP_FIXEDHDR_LEN 236

typedef struct {
    uint8_t op;
    uint8_t htype;
    uint8_t hlen;
    uint8_t hops;
    uint32_t xid;
    uint16_t secs;
    uint16_t flags;
    uint32_t ciaddr;
    uint32_t yiaddr;
    uint32_t siaddr;
    uint32_t giaddr;
    
#define DHCP_CHADDR_SIZE 16
    unsigned char chaddr[DHCP_CHADDR_SIZE];
#define DHCP_SNAME_SIZE 64
    unsigned char sname[DHCP_SNAME_SIZE];
#define DHCP_FILE_SIZE 128
    unsigned char file[DHCP_FILE_SIZE];

    /* options go here. */
} dhcphdr;

/* dhcp option structure for internal
 * dhcp object representation. */

typedef struct {
    unsigned char tag;
    unsigned char len;
    unsigned char *data;
} dhcp_option_t;

/* DHCP object *
 * for packet assembly/disassembly. */

typedef struct {
    dhcphdr fixedheader;          /* DHCP fixed header. */
    uint32_t magic_cookie;        /* Magic cookie. */
    list_t *options;              /* Option list. */
    list_t *options_seek;         /* Seek pointer to option list. */
} dhcp_obj;

/* 
 * Other packet objects.
 * 
 * We use these to create a single interface
 * for creating custom packeting.
 * 
 */

typedef struct {
    struct eth_hdr header;
} eth_obj;

typedef struct {
    struct ip_hdr header;
} ip_obj;

typedef struct {
    struct arp_hdr header;
    struct arp_ethip arp_data;
} arp_obj;

typedef struct {
    struct icmp_hdr icmp_header;
    union  icmp_msg icmp_msg;
} icmp_obj;

typedef struct {
    struct udp_hdr header;
} udp_obj;

/* interface control object:
 * we store dnet's interface handle, 
 * and interface_entry structure. 
 * We need the entry structure
 */

typedef struct {
    intf_t *interface_handle;
    struct intf_entry *interface_entry;
} interface_control_t;


/* 
 * DHCP raw network object.
 * 
 * We do all our network transmission
 * through this data structure.
 * 
 */

typedef struct {
    
    /* we should not interface directly with any of these members
     * above the rawnet layer: FIXME -- unfortunately we do. */

    pcap_t *pcap;                  /* packet capturing device (pcap). */
    int pcap_fd;                   /* file descriptor we can select on for pcap. */
    eth_t *eth;                    /* ethernet device handler (libdnet). */

    uint16_t src_port, dst_port;   /* udp ports for writing packets. */
    char *device;                  /* device name. */

    char *packet_data;             /* packet data. */
    int packet_len;                /* total length of packet. */

    struct timeval tstamp;         /* timestamp. */

    eth_addr_t hw_addr;            /* our hardware address. */
    eth_addr_t *fake_hw_addr;      /* our fake hardware address. */
    ip_addr_t ip_addr;             /* our ip address. */
    int promiscuous;               /* run in promiscuous mode flag. */
    interface_control_t 
                  *intf_handle;    /* interface control handle. */

    /* 
     * Data objects
     * 
     * After reception of packets on the wire
     * they're aligned and placed into these
     * objects.
     * 
     */

    unsigned char type;         /* Type of data:
                                 * ARP, ICMP DHCP */
    char *pcap_filter;          /* low level pcap filter. */

    /* Packet data. */

    eth_obj   *ether_p;
    ip_obj    *ip_p;
    arp_obj   *arp_p;
    icmp_obj  *icmp_p;
    udp_obj   *udp_p;
    dhcp_obj  *dhcp_p;

} rawnet_t;

typedef struct {
    char *conf_file;
    char *interface; /* points to dhcp_client_control->interface */
    
    /* specific configurations */

    unsigned char options[MAX_OPTIONS_HANDLED]; /* which options should we handle. */
} client_conf_t;

/* Cache data types. */

typedef struct {
    char *name;
    char *value;
} cache_entry_t;

typedef struct {
    char *interface; /* points to dhcp_client_control->interface */
    list_t *vars;
} client_cache_t;

/* Client data object. */

typedef struct {
    
    rawnet_t *rawnet; /* raw network handle */
    
    client_cache_t *cache;       /* cache */
    client_conf_t *conf;         /* configuration. */
    
    uint32_t xid;                /* unique xid */
    time_t started;              /* the amount of milliseconds since we started. */
    time_t secs;                 /* secs -- used to save secs value. */
    uint32_t renewal_time;       /* when to renew lease (secs). */

    char *interface;                        /* interface name. */
    unsigned char *class_id, *client_id;    /* client_id, class_id */
    int state;                              /* our current state. */
    int discover_offer_retries;             /* counter for retries on discover_offer */
} dhcp_client_control_t;

/* Option print handler structure. */

typedef struct {
    char *prefix;
    void (*handle_option)(const unsigned char *data,
                          int len, int print_val);
} option_handler;

/* * * * * * * 
 * Constants *
 * * * * * * */

/* Syslog options. */

#define LOG_OPT LOG_PID|LOG_NDELAY /* Options  */
#define LOG_FACILITY LOG_DAEMON    /* Facility */
#define ERROR_LEVEL LOG_ERR        /* Error level */
#define INFO_LEVEL LOG_INFO        /* Normal level */

/* DHCP Type messages */

# define DHCP_DISCOVER_TM 1
# define DHCP_OFFER_TM    2
# define DHCP_REQUEST_TM  3
# define DHCP_DECLINE_TM  4
# define DHCP_DHCPACK_TM  5
# define DHCP_DHCPNAK_TM  6
# define DHCP_RELEASE_TM  7

/* BOOTP Types */

#define DHCP_BOOTP_REQUEST 1
#define DHCP_BOOTP_REPLY 2

/* Flags */

#define DHCP_BROADCAST_FLAG 0x8000

/* Option tags. */

# define TAG_PAD               0
# define TAG_END               255

# define TAG_DHCP_SUBNET_MASK            1
# define TAG_DHCP_HOSTNAME              12
# define TAG_DHCP_REQUEST_IP_ADDRESS    50
# define TAG_DHCP_OPTION_OVERLOAD       52
# define TAG_DHCP_MESSAGE_TYPE          53
# define TAG_DHCP_SERVER_IDENTIFIER     54
# define TAG_PARAMETERS                 55
# define TAG_MAX_DHCP_SIZE              57
# define TAG_DHCP_CLASS_ID              60
# define TAG_DHCP_CLIENT_ID             61

/* DHCP overload option values. */

#define DHCP_OVERLOAD_FILE  1
#define DHCP_OVERLOAD_SNAME 2
#define DHCP_OVERLOAD_BOTH  3
    
/* NetBIOS Scope Node Types */

# define NETBIOS_B_NODE 0x01
# define NETBIOS_P_NODE 0x02
# define NETBIOS_M_NODE 0x04
# define NETBIOS_H_NODE 0x06

/* DHCP states
 * We use these internally
 * for the state machine. */

#define STATE_FATAL_ERROR    -1
#define STATE_DISCOVER_OFFER 0
#define STATE_REQUEST_ACK    1
#define STATE_SETUP          3
#define STATE_WAIT           4
#define STATE_RELEASE        5

/* Parse types. */

#define PARSE_SINGLE_STRING   1
#define PARSE_DOUBLE_STRINGS  2

/* * * * * * * * * * * *
 * Function prototypes *
 * * * * * * * * * * * */

/* cache manipulation. */

extern client_cache_t *create_client_cache(dhcp_client_control_t *dc);
extern void client_cache_destroy(client_cache_t *cc);
extern int load_client_cache(client_cache_t *dc, unsigned char use_tmp);
extern int client_cache_is_empty(client_cache_t *cc);
extern void purge_cache(client_cache_t *cc);
extern int client_cache_dump_options(client_cache_t *cc, list_t *options);
extern list_t *client_cache_load_option_network_list(client_cache_t *cc, 
                                                     unsigned char use_temp);
extern list_t *client_cache_load_option_string_list(client_cache_t *cc,
                                                    unsigned char use_tmp);

extern void client_cache_update(client_cache_t *cc);
extern void client_cache_delete_cache(client_cache_t *cc);
extern void client_cache_delete_tmp_cache(client_cache_t *cc);

/* cache object manipulation. */

extern cache_entry_t *make_cache_entry(const char *name, const char *value);
extern void destroy_cache_entry(cache_entry_t *cache);
extern void cache_entry_purge_list(list_t *cache_list);

/* packet building routines. */

extern void build_dhcp_discover(rawnet_t *net, uint32_t xid, time_t secs, list_t *options);
extern void build_dhcp_request(rawnet_t *net, uint32_t xid, time_t secs, list_t *options);
extern void build_arp_reply_broadcast(rawnet_t *net, uint32_t source_addr, 
                                      eth_addr_t source_hw_addr);

extern void build_arp_reply(rawnet_t *net,
                            uint32_t source_addr,
                            uint32_t dest_addr,
                            eth_addr_t source_hw_addr,
                            eth_addr_t dest_hw_addr);

extern void build_arp_request(rawnet_t *net,
                              uint32_t source_addr,
                              uint32_t dest_addr,
                              eth_addr_t source_hw_addr);

extern void build_unarp(rawnet_t *net, uint32_t source_addr, 
                        eth_addr_t source_hw_addr);
extern void build_icmp_mask_request(rawnet_t *net, uint32_t id, uint32_t seq);
extern void build_icmp_mask_reply(rawnet_t *net, uint32_t id, uint32_t seq, uint32_t subnet_mask);
extern void build_icmp_echo_request(rawnet_t *net, ip_addr_t source_addr, ip_addr_t dest_addr,
                                    eth_addr_t source_mac, eth_addr_t dest_mac, uint32_t id, 
                                    uint32_t seq);

/* ICMP discovery routines. */
extern int icmp_subnet_mask_discovery(rawnet_t *net, int retries, uint32_t *subnet_mask);
extern list_t *icmp_rtt_discovery(rawnet_t *net, list_t *addresses);

/* ARP discovery routines. */
extern int arp_discover_hardware_address(rawnet_t *net, int retries, uint32_t address, eth_addr_t *mac_addr);

/* routing table routines. */

extern int route_find(rawnet_t *net, ip_addr_t addr, eth_addr_t *dest_mac);

/* Daemon routines. */

extern void go_background(char *dir);
extern void setup_interrupt_handlers(void);

/* Replacement vsnprintf, snprintf that work the way we want them to. */

extern int snprintf (char *str, size_t count, const char *fmt, ...);
extern int vsnprintf (char *str, size_t count, const char *fmt, va_list arg);

/* DHCP obj routines. */

extern dhcp_obj *dhcp_create(void);
extern void dhcp_purge(dhcp_obj *dhcp);
extern void dhcp_purge_option_list(list_t *options); /* only use on network list. */
extern void dhcp_destroy(dhcp_obj *dhcp);
extern int dhcp_read_packet_image(dhcp_obj *dhcp, const unsigned char *dhcp_packet, int len);
extern void dhcp_write_packet_image(dhcp_obj *dhcp, unsigned char *packet);

extern unsigned char dhcp_get_op(dhcp_obj *dhcp);
extern unsigned char dhcp_get_hlen(dhcp_obj *dhcp);
extern unsigned char dhcp_get_htype(dhcp_obj *dhcp);
extern unsigned char dhcp_get_hops(dhcp_obj *dhcp);
extern uint32_t dhcp_get_xid(dhcp_obj *dhcp);
extern uint16_t dhcp_get_secs(dhcp_obj *dhcp);
extern uint16_t dhcp_get_flags(dhcp_obj *dhcp);
extern uint32_t dhcp_get_ciaddr(dhcp_obj *dhcp);
extern uint32_t dhcp_get_yiaddr(dhcp_obj *dhcp);
extern uint32_t dhcp_get_siaddr(dhcp_obj *dhcp);
extern uint32_t dhcp_get_giaddr(dhcp_obj *dhcp);
extern unsigned char *dhcp_get_chaddr(dhcp_obj *dhcp);
extern unsigned char *dhcp_get_sname(dhcp_obj *dhcp);
extern unsigned char *dhcp_get_filename(dhcp_obj *dhcp);
extern uint32_t dhcp_get_magic_cookie(dhcp_obj *dhcp);
    
extern void dhcp_set_op(dhcp_obj *dhcp, unsigned char op);
extern void dhcp_set_htype(dhcp_obj *dhcp, unsigned char htype);
extern void dhcp_set_hlen(dhcp_obj *dhcp, unsigned char hlen);
extern void dhcp_set_hops(dhcp_obj *dhcp, unsigned char hops);
extern void dhcp_set_xid(dhcp_obj *dhcp, uint32_t xid);
extern void dhcp_set_secs(dhcp_obj *dhcp, uint16_t secs);
extern void dhcp_set_flag_broadcast(dhcp_obj *dhcp);
extern void dhcp_set_ciaddr(dhcp_obj *dhcp, uint32_t ciaddr);
extern void dhcp_set_yiaddr(dhcp_obj *dhcp, uint32_t yiaddr);
extern void dhcp_set_siaddr(dhcp_obj *dhcp, uint32_t siaddr);
extern void dhcp_set_giaddr(dhcp_obj *dhcp, uint32_t giaddr);
extern void dhcp_set_sname(dhcp_obj *dhcp, unsigned char *sname);
extern void dhcp_set_filename(dhcp_obj *dhcp, unsigned char *file);
extern void dhcp_set_magic_cookie(dhcp_obj *dhcp);
extern void dhcp_set_options(dhcp_obj *dhcp, list_t *options);
extern void dhcp_set_chaddr(dhcp_obj *dhcp, unsigned char *chaddr, int len);

extern void dhcp_clear_sname(dhcp_obj *dhcp);
extern void dhcp_clear_filename(dhcp_obj *dhcp);
extern void dhcp_clear_chaddr(dhcp_obj *dhcp);

extern void dhcp_reset_option_seek(dhcp_obj *dhcp);
extern dhcp_option_t *dhcp_get_next_option(dhcp_obj *dhcp);
extern dhcp_option_t *create_dhcp_option(const unsigned char *opt_data, 
                                       unsigned char len, unsigned char tag);
extern void destroy_dhcp_option(dhcp_option_t *option);
extern uint16_t dhcp_get_options_len(list_t *options);
extern int dhcp_valid_magic_cookie(dhcp_obj *dhcp);
extern int dhcp_is_type(dhcp_obj *dhcp, unsigned char type);
extern int dhcp_option_is_valid(unsigned char tag, unsigned char tag_len);
extern int dhcp_have_atleast_requested_options(dhcp_obj *dhcp, unsigned char *option);
extern int dhcp_have_option(dhcp_obj *dhcp, unsigned char tag);
extern int dhcp_is_file_overload(dhcp_obj *dhcp);
extern int dhcp_is_sname_overload(dhcp_obj *dhcp);

extern dhcp_option_t *dhcp_build_parameter_request_list_option(unsigned char *requested_options);
extern dhcp_option_t *dhcp_build_max_message_size_option(uint16_t max);
extern dhcp_option_t *dhcp_build_message_type(unsigned char type);
extern dhcp_option_t *dhcp_build_class_id(unsigned char *class_id);
extern dhcp_option_t *dhcp_build_client_id(unsigned char *client_id, unsigned char len);
extern dhcp_option_t *dhcp_build_requested_ip_address(uint32_t address);
extern dhcp_option_t *dhcp_build_server_identifier(uint32_t address);
extern dhcp_option_t *dhcp_build_hostname(char *hostname);
extern uint32_t dhcp_gen_xid(void);

/* Ether obj routines. */

extern eth_obj *eth_create(void);
extern void eth_destroy(eth_obj *eth);
extern int eth_read_packet_image(eth_obj *eth, const unsigned char *packet, int len);
extern void eth_write_packet_image(eth_obj *eth, unsigned char *packet);

extern unsigned char *eth_get_src_address(eth_obj *eth);
extern unsigned char *eth_get_dst_address(eth_obj *eth);
extern uint16_t eth_get_type(eth_obj *eth);

extern void eth_set_src_address(eth_obj *eth, eth_addr_t addr);
extern void eth_set_dst_address(eth_obj *eth, eth_addr_t addr);
extern void eth_set_type(eth_obj *eth, uint16_t type);

/* IP obj routines. */

extern ip_obj *ip_create(void);
extern void ip_destroy(ip_obj *ip);
extern int ip_read_packet_image(ip_obj *ip, const unsigned char *packet, int len);
extern void ip_write_packet_image(ip_obj *ip, unsigned char *packet);

extern uint32_t ip_get_source_addr(ip_obj *ip);
extern uint32_t ip_get_dest_addr(ip_obj *ip);
extern unsigned char ip_get_hl(ip_obj *ip);
extern uint8_t ip_get_proto(ip_obj *ip);

extern void ip_set_hl(ip_obj *ip, uint16_t header_len);
extern void ip_set_tos(ip_obj *ip, unsigned char ip_tos);
extern void ip_set_len(ip_obj *ip, uint16_t ip_len);
extern void ip_set_id(ip_obj *ip, uint16_t ip_id);
extern void ip_set_off(ip_obj *ip, uint16_t ip_off);
extern void ip_set_ttl(ip_obj *ip, uint8_t ip_ttl);
extern void ip_set_proto(ip_obj *ip, unsigned char ip_proto);
extern void ip_set_source_addr(ip_obj *ip, uint32_t ip_src);
extern void ip_set_dest_addr(ip_obj *ip, uint32_t ip_src);

/* ARP obj routines. */

extern arp_obj *arp_create(void);
extern void arp_destroy(arp_obj *arp);
extern void arp_write_packet_image(arp_obj *arp, unsigned char *packet);

extern void arp_set_hardware_type(arp_obj *arp, uint16_t hdr_addr);
extern void arp_set_protocol_type(arp_obj *arp, uint16_t pro);
extern void arp_set_hardware_len(arp_obj *arp, uint8_t len);
extern void arp_set_protocol_len(arp_obj *arp, uint8_t len);
extern void arp_set_op(arp_obj *arp, uint16_t op);
extern void arp_set_sender_hardware_address(arp_obj *arp, eth_addr_t addr);
extern void arp_set_sender_protocol_address(arp_obj *arp, uint32_t addr);
extern void arp_set_target_hardware_address(arp_obj *arp, eth_addr_t addr);
extern void arp_set_target_protocol_address(arp_obj *arp, uint32_t addr);

extern uint16_t arp_get_op(arp_obj *arp);
extern uint16_t arp_get_hardware_type(arp_obj *arp);
extern uint16_t arp_get_protocol_type(arp_obj *arp);
extern uint8_t arp_get_hardware_len(arp_obj *arp);
extern uint8_t arp_get_protocol_len(arp_obj *arp);
extern unsigned char *arp_get_sender_hardware_address(arp_obj *arp);
extern unsigned char *arp_get_sender_protocol_address(arp_obj *arp);

extern int arp_read_packet_image(arp_obj *arp, const unsigned char *packet, int len);

/* UDP obj routines. */

extern udp_obj *udp_create(void);
extern void udp_destroy(udp_obj *udp);
extern int udp_read_packet_image(udp_obj *udp, const unsigned char *packet, int len);
extern void udp_write_packet_image(udp_obj *udp, unsigned char *packet);

extern uint16_t udp_get_src_port(udp_obj *udp);
extern uint16_t udp_get_dst_port(udp_obj *udp);
extern uint16_t udp_get_len(udp_obj *udp);

extern void udp_set_src_port(udp_obj *udp, uint16_t src_port);
extern void udp_set_dst_port(udp_obj *udp, uint16_t dst_port);
extern void udp_set_len(udp_obj *udp, uint16_t len);
extern void udp_set_cksum(udp_obj *udp, uint16_t cksum);

/* ICMP obj routines. */

extern icmp_obj *icmp_create(void);
extern void icmp_destroy(icmp_obj *icmp);

extern void icmp_set_type(icmp_obj *icmp, uint8_t type);
extern void icmp_set_code(icmp_obj *icmp, uint8_t code);
extern void icmp_clear_cksum(icmp_obj *icmp);
extern void icmp_mask_set_id(icmp_obj *icmp, uint32_t id);
extern void icmp_mask_set_seq(icmp_obj *icmp, uint32_t seq);
extern void icmp_mask_set_mask(icmp_obj *icmp, uint32_t mask);
extern void icmp_echo_set_id(icmp_obj *icmp, uint32_t id);
extern void icmp_echo_set_seq(icmp_obj *icmp, uint32_t seq);

extern uint8_t icmp_get_type(icmp_obj *icmp);
extern uint32_t icmp_mask_get_mask(icmp_obj *icmp_p);
extern void icmp_write_packet_image(icmp_obj *icmp, unsigned char *packet);
extern int icmp_read_packet_image(icmp_obj *icmp, const unsigned char *packet, int len);

/* Alignment functions. */

extern void align_eth(const unsigned char *data, struct eth_hdr *eth);
extern void align_ip(const unsigned char *data, struct ip_hdr *iphdr);
extern void align_udp(const unsigned char *data, struct udp_hdr *udp);
extern void align_dhcphdr(const unsigned char *data, dhcphdr *dhcp);
extern void align_icmp(const unsigned char *data, struct icmp_hdr *header);
extern void align_icmp_mask(const unsigned char *data, struct icmp_msg_mask *header);
extern void align_icmp_echo(const unsigned char *data, struct icmp_msg_echo *header);
extern void align_arp_header(const unsigned char *data, struct arp_hdr *header);
extern void align_arp_data(const unsigned char *data, struct arp_ethip *arp_data);

/* Linked List routines. */

extern list_t *add_to_list(list_t *head, void *datum);
extern list_t *add_to_end_of_list(list_t *head, void *datum);
extern list_t *remove_from_list(list_t *head, void *datum);
extern void purge_list(list_t *head, void (*purge_func)(void *d));
extern void purge_list_internal(void *head);
extern list_t *join_lists(list_t *first_list, list_t *second_list);

/* Logging functions. */

extern void init_log(const char *s);
extern void close_log(void);
extern void error_log(char *msg);
extern void info_log(char *msg);

/* dhcp client control */

extern dhcp_client_control_t *create_dhcp_client_control(char *interface, int promiscuous);
extern void destroy_dhcp_client_control(dhcp_client_control_t *dc);
extern void dhcp_client_update_secs(dhcp_client_control_t *dc);
extern void dhcp_client_reset_secs(dhcp_client_control_t *dc);
extern void dhcp_control_use_fake_hw_addr(dhcp_client_control_t *dc, 
                                          char *fake_hw_addr);
extern dhcp_client_control_t *create_dhcp_client_control_dummy(char *interface);
extern void dhcp_client_interface_down(dhcp_client_control_t *dc);

/* dhcp client conf */

extern client_conf_t *create_client_conf(dhcp_client_control_t *dc);
extern void client_conf_destroy(client_conf_t *cc);
extern int load_client_conf(client_conf_t *cc);

/* interface control */

extern interface_control_t *create_interface_control(char *name);
extern int interface_is_up(interface_control_t *ic);
extern int interface_up(interface_control_t *ic, uint32_t addr, uint32_t netmask, int mtu);
extern int interface_down(interface_control_t *ic);
extern void destroy_interface_control(interface_control_t *ic);
extern int interface_get_ip_addr(interface_control_t *ic, uint32_t *addr);
extern list_t *interface_get_active_interfaces(void);
extern list_t *interface_get_inactive_interfaces(void);

/* client states */

extern int client_discover_offer(dhcp_client_control_t *dc);
extern int client_request_ack(dhcp_client_control_t *dc);
extern int client_wait(dhcp_client_control_t *dc);
extern int client_release(dhcp_client_control_t *dc);
extern int client_setup(dhcp_client_control_t *dc);

/* Global vars. */

extern int interactive;
extern char *workdirectory;
extern int want_shutdown, want_hup;
extern char *dhcp_conf_options_strings[];
extern option_handler option_handlers[];
extern char *dhcp_options_strings[];

#endif /* DHCP_AGENT_H */
