/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-client-cache.c,v 1.13 2002/06/22 18:11:25 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * Client cache object:
 * 
 * We use this to load cached DHCP options, save them, and conversion with
 * dhcp-convert are done here to network or serialized state.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-files.h>
#include <dhcp-parser.h>
#include <dhcp-varfile.h>
#include <dhcp-convert.h>

/* constructor. */
client_cache_t *create_client_cache(dhcp_client_control_t *dc)
{
    client_cache_t *cc;
    
    cc = xmalloc(sizeof(client_cache_t));
    cc->interface = dc->interface; /* don't copy just point. */
    cc->vars = NULL;
    
    return cc;
}

/* purges list in cache object. */
void purge_cache(client_cache_t *cc)
{
    list_t *list_ptr, *tmp;
    cache_entry_t *ce;
    
    list_ptr = cc->vars;
    while(list_ptr)  {
        
        tmp = list_ptr;
        list_ptr = list_ptr->next;
        
        ce = tmp->data;
        destroy_cache_entry(ce);
        xfree(tmp);
    }

    cc->vars = NULL;
    return;
}

/* destructor. */
void client_cache_destroy(client_cache_t *cc)
{
    if(cc->vars)
        purge_cache(cc);
    
    xfree(cc);
    return;
}

/* read next cache entry from file. */
static cache_entry_t *read_next_cache_var(varfile_t *varfile)
{
    cache_entry_t *cache_entry;
    int retval;

    retval = varfile_get_next_var(varfile);
    switch(retval) {

      case VARFILE_PARSE_ERROR:
          fatal_error("client-cache: parse error in cache: %s : line %d", 
                      varfile_get_filename(varfile), varfile_get_lineno(varfile));

    case VARFILE_EOF:
        return NULL;

    case VARFILE_SUCCESS:
        cache_entry = make_cache_entry(varfile_get_name(varfile), varfile_get_val(varfile));
        return cache_entry;

    default:
        fatal_error("client-cache: read_next_cache_var: internal error -- please report this.");
    }

    /* we never get here. */
    return NULL;
}

/* Load up cache from filename. */
static int load_client_cache_proc(client_cache_t *cc, char *filename)
{
    varfile_t *varfile;
    cache_entry_t *cache_entry;

    varfile = create_varfile(filename, VARFILE_VARVAL_MODE);
    if(varfile == NULL)
        return -1;
    
    /* Read up any vars. */

    while((cache_entry = read_next_cache_var(varfile)) != NULL)
        cc->vars = add_to_list(cc->vars, cache_entry);

    destroy_varfile(varfile);

    return 0;
}

/* filename creation routines. */
static char *get_fname(client_cache_t *cc)
{
    return(splice_string(cc->interface, ".cache"));
}

static char *get_fname_tmp(client_cache_t *cc)
{
    return(splice_string(cc->interface, ".cache.tmp"));
}

/* load cache (interface) */
int load_client_cache(client_cache_t *cc, unsigned char use_tmp)
{
    char *fname;
    int retval;
    
    purge_cache(cc);
    
    if(use_tmp)
        fname = get_fname_tmp(cc);
    else
        fname = get_fname(cc);

    retval = load_client_cache_proc(cc, fname);
    xfree(fname);
    
    return retval;
}

/* check to see if cache is empty:
 * as long as the cache file exists
 * we're always guaranteed a cache
 * except on filesystem corruption.
 * in that case we can't do much except
 * complain. using file_exists() is
 * as good as it gets.
 */
int client_cache_is_empty(client_cache_t *cc)
{
    int retval = 0;
    char *fname;
    
    fname = get_fname(cc);
    if(file_exists(fname))
        retval = 1;
    xfree(fname);
    
    return retval;
}

/* work horse routine to convert from serialized state to network datum. */
static void *convert_to_network(unsigned char tag, char *string)
{
    void *data;
    dhcp_option_t *opt;
    unsigned char len;
    
    if(option_convert_handlers[tag].serialize_to_network == NULL)
        return NULL;
    
    data = option_convert_handlers[tag].serialize_to_network(string, &len);

    if(data == NULL)  {
        warn_message("option: %s may be malformed. I'll ignore and carry on.",
                      dhcp_options_strings[tag]);
        return NULL;
    }

    opt = create_dhcp_option(data, len, tag);
    xfree(data);
    return opt;
}

/* 
 * Convert options from cache serialized form to whatever form convert_func, returns.
 * If convert_func is NULL then just return in serialized form (sysconf needs it that way).
 * 
 */
static list_t *get_cache_options_list_proc(client_cache_t *cc, 
                                           void* (*convert_func)(unsigned char tag, char *string))
{
    list_t *options = NULL;
    list_t *cache;
    cache_entry_t *ce, *copy;
    unsigned char i;
    void *data;

    for(cache = cc->vars;cache;cache = cache->next)  {
        
        ce = cache->data;
        
        for(i = 0;i < MAX_OPTIONS_HANDLED;i++)  {

            if(string_matches(dhcp_options_strings[i], ce->name))  {

                if(convert_func)  {
                    data = convert_func(i, ce->value);
                    if(data == NULL)
                        break; /* break out of for loop since we have a match and we're
                                * done even if we couldn't get the data. */
                    options = add_to_list(options, data);
                    break;
                } else {
                    copy = make_cache_entry(ce->name, ce->value);
                    options = add_to_list(options, copy);
                    break;
                }
            }
        }
    }
        
    return options;
}

/* dump options into file. here we take data in network form and dump them in serialized form.
 * we always do this to the temp file.
 */
int client_cache_dump_options(client_cache_t *cc,
                               list_t *options)
{
    FILE *fp;
    list_t *list_ptr;
    char *name;
    char *val;
    dhcp_option_t *opt;

    name = get_fname_tmp(cc);
    fp = file_create_and_truncate_safe(name, "a");
    xfree(name);

    if(fp == NULL)
        return -1;

    for(list_ptr = options;list_ptr;list_ptr = list_ptr->next)  {
        
        /* Run coversion routines to store internally in 
         * serialized (string) state. */
        
        opt = list_ptr->data;

        if(option_convert_handlers[opt->tag].serialize == NULL)
            continue;

        name = dhcp_options_strings[opt->tag];
        val = option_convert_handlers[opt->tag].serialize(opt->data, opt->len);

        /* 
         * the dhcp data object cannot validate
         * at the level we can, so check for a NULL value.
         * 
         */

        if(val == NULL)  {
            warn_message("option %s appear malformed. skipping..");
            continue;
        }

        fprintf(fp, "\"%s\"=%s\n", name, val);
        xfree(val);
    }

    fclose(fp);
    return 0;
}

/* work horse load options routine. */
static list_t *client_cache_load_options_list_proc(client_cache_t *cc, unsigned char use_tmp,
                                                  void* (*convert_func)(unsigned char tag, char *string))
{
    
    list_t *options;

    if(load_client_cache(cc, use_tmp) < 0)
        return NULL;

    options = get_cache_options_list_proc(cc, convert_func);
    purge_cache(cc);

    return options;
}

/* load options and return network list. */
list_t *client_cache_load_option_network_list(client_cache_t *cc, unsigned char use_tmp)
{
    return(client_cache_load_options_list_proc(cc, use_tmp, convert_to_network));
}
/* load options and return serialized list. */
list_t *client_cache_load_option_string_list(client_cache_t *cc, unsigned char use_tmp)
{
    return(client_cache_load_options_list_proc(cc, use_tmp, NULL));
}

/* update cache: move temp cache over real cache. */
void client_cache_update(client_cache_t *cc)
{
    char *fname_tmp, *fname;
    
    fname_tmp = get_fname_tmp(cc);
    fname = get_fname(cc);
    
    move_file(fname_tmp, fname);

    xfree(fname);
    xfree(fname_tmp);

    return;
}

/* delete old temporary cache. */
void client_cache_delete_tmp_cache(client_cache_t *cc)
{
    char *fname = get_fname_tmp(cc);

    delete_file(fname);
    xfree(fname);
    
    return;
}

/* delete cache. */
void client_cache_delete_cache(client_cache_t *cc)
{
    char *fname = get_fname(cc);

    delete_file(fname);
    xfree(fname);
    
    return;
}
