/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-client-conf.c,v 1.12 2002/06/22 18:11:25 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-files.h>
#include <dhcp-parser.h>
#include <dhcp-varfile.h>

/* Basic network configuration.
 * Here we keep the options we want by default. */

static const int default_config_options[] = { 
        1,  /* Subnet Mask */
        3,  /* Routers. */
        6,  /* Domain Name Servers. */
        15, /* Domain Name. */
        50, /* IP Address lease time. */
        54, /* Server Identifier. */
        58, /* Renewal time. */
};


static const int default_config_options_len = 7;

static void client_conf_reset_options(client_conf_t *cc)
{
    memset(cc->options, 0, sizeof(cc->options));
}

client_conf_t *create_client_conf(dhcp_client_control_t *dc)
{
    client_conf_t *cc;
    
    cc = xmalloc(sizeof(client_conf_t));
    client_conf_reset_options(cc);
    cc->interface = dc->interface;
    return cc;
}

void client_conf_destroy(client_conf_t *cc)
{
     xfree(cc);
}

static char *get_conf_options_fname(client_conf_t *cc)
{
    return(splice_string(cc->interface, ".opts"));
}

static int dump_config_options(client_conf_t *cc, char *fname)
{
    int i;
    FILE *fp;
    
    fp = file_open_or_create_safe(fname, "a+");

    if(fp == NULL)
        return -1;

    for(i = 0;i < default_config_options_len; i++)  {
        
        if(fprintf(fp, "\"%s\"\n",dhcp_options_strings[default_config_options[i]]) < 0) {
            warn_message("could not dump options to file. I'm going to continue anyway.");
            fclose(fp);
            return -1;
        }
    }
    
    fclose(fp);
    return 0;
}

static void use_default_options(client_conf_t *cc)
{
    int i;

    for(i = 0;i < default_config_options_len; i++)  {
        cc->options[default_config_options[i]] = 1;
    }
    
    return;
}

static int client_conf_load_options(client_conf_t *cc, char *fname)
{
    varfile_t *varfile;
    int i;
    unsigned char match;
    int retval;

    varfile = create_varfile(fname, VARFILE_VARNAMES_MODE);

    if(varfile == NULL)
        return -1;
    
    while(1) {
        
        retval = varfile_get_next_var(varfile);

        switch(retval) {

          case VARFILE_SUCCESS:

              match = 0;
              for(i = 0;i < MAX_OPTIONS_HANDLED;i++)   {

                  /* as long as no error was returned from file_parse_string
                   * we can just do a file_get_var_name() safely. */

                  if(string_matches(dhcp_options_strings[i],
                                    varfile_get_name(varfile)))  {
                      cc->options[i] = 1;
                      match = 1;
                      break;
                  }
                  
              }

        if(match != 1)
            warn_message("encountered illegal DHCP option in %s: %s : %d", fname, varfile_get_name(varfile),
                         varfile_get_lineno(varfile));
        break;

        case VARFILE_EOF:
            destroy_varfile(varfile);
            return 0;

        case VARFILE_PARSE_ERROR:
            fatal_error("client-conf: parse error in opts file: %s : %d", 
                        varfile_get_filename(varfile), varfile_get_lineno(varfile));
        default:
            fatal_error("client-conf: load_options: internal error -- please report this.");
        }
    }
}

/* load configuration file or call applicable default setter,
   also call dump_defaults if file does not exist */

static int load_client_conf_proc(char *fname, 
                                 client_conf_t *cc, int (*load_conf)(client_conf_t *cc, char *fname), 
                                 void (*use_defaults)(client_conf_t *cc),
                                 int (*dump_defaults)(client_conf_t *cc, char *fname))
{
    int retval;

    if(!file_exists(fname))  {
        /* it doesn't exist. strap default. */
        use_defaults(cc);
        /* and dump to file. */
        retval = dump_config_options(cc, fname);
    } else {
        /* read conf data from file. */
        retval = load_conf(cc, fname);
    }

    return retval;
}

int load_client_conf(client_conf_t *cc)
{
    
    char *fname;
    int retval;

    /* load up options file. */

    client_conf_reset_options(cc);
    /* check for options file. */
    fname = get_conf_options_fname(cc);
    retval = load_client_conf_proc(fname, cc, client_conf_load_options, use_default_options, dump_config_options);
    xfree(fname);

    return(retval);
}
