/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-client-control.c,v 1.22 2002/06/22 01:47:57 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-globconf.h>
#include <dhcp-net.h>

/* Utility routines to update counters and time stamps. */

/* Update seconds field for dhcp discover packets. */
void dhcp_client_update_secs(dhcp_client_control_t *dc)
{
    dc->secs = (time(NULL) - dc->started);
    return;
}

/* reset started field for when we revert back to discover again. */
void dhcp_client_reset_secs(dhcp_client_control_t *dc)
{
    dc->started = time(NULL);
    return;
}

/* dhcp reset discover offer retries. */
void dhcp_client_reset_discover_offer_retries(dhcp_client_control_t *dc)
{
    dc->discover_offer_retries = 0;
}

/* dhcp increment and check whether we should give up on discover offer retries. */
int dhcp_client_discover_offer_can_retry(dhcp_client_control_t *dc)
{
    uint32_t *discover_offer_retries = glob_conf_get_val(CLIENT_DHCP_RETRIES);
    dc->discover_offer_retries++;

    if(dc->discover_offer_retries > *(discover_offer_retries))
        return 1;

    return 0;
}
/* create client id from a string in the format of xx:xx:xx:xx:xx:xx */
static unsigned char *create_fake_client_id(unsigned char *mac_string)
{
    unsigned char *s;
    unsigned char mac_add[ETH_ADDR_LEN];

    s = xmalloc(1 + ETH_ADDR_LEN);
    *s = DLT_EN10MB;
    s++;
    
    if(hex_string_to_value(mac_string, mac_add))
        fatal_error("malformed mac_string for fake_client_id. please check mac address.");
        
    memcpy(s, mac_add, ETH_ADDR_LEN);
    s--;

    return s;
}

/* create a proper client id from the interface address passed. */
static unsigned char *create_client_id(eth_addr_t interface_address)
{
    unsigned char *s;
    
    s = xmalloc(1 + ETH_ADDR_LEN);
    *s = DLT_EN10MB;
    s++;
    memcpy(s, interface_address.data, ETH_ADDR_LEN);
    s--;
    
    return s;
}

/* dummy routine. we use this to manipulate conf/cache. */
dhcp_client_control_t *create_dhcp_client_control_dummy(char *interface)
{
    dhcp_client_control_t *dc;
    
    dc = xmalloc(sizeof(dhcp_client_control_t));
    memset(dc, 0, sizeof(dhcp_client_control_t));
    dc->interface = strdup(interface);
    
    return dc;
}

static void update_dhcp_xid(dhcp_client_control_t *dc)
{
    dc->xid = dhcp_gen_xid();
    return;
}

/* create client control object */
dhcp_client_control_t *create_dhcp_client_control(char *interface, int promiscuous)
{
    dhcp_client_control_t *dc;
    int dport, sport;
    char filter_buff[GENERIC_BUFFSIZE];
    eth_addr_t interface_addr;

#ifdef HAVE_UNAME
    struct utsname utsname;
#endif /* HAVE_UNAME */

    dc = xcalloc(sizeof(dhcp_client_control_t));
    
    dc->state = 0;
    dc->interface = strdup(interface);
    dc->conf = create_client_conf(dc);
    dc->cache = create_client_cache(dc);

    if((load_client_conf(dc->conf) < 0) || (load_client_cache(dc->cache, 0) < 0))  {
        destroy_dhcp_client_control(dc);
        return NULL;
    }

    /* Get port numbers from services db. */

    dport = port_for_service("bootps", "udp");
    sport = port_for_service("bootpc", "udp");

    dport = ntohs(dport);
    sport = ntohs(sport);
    
    if(dport == -1 || sport == -1)  {

        warn_message("could not lookup dhcp services in service db: %s", strerror(errno));
        warn_message("will use reasonable default.");

        sport = BOOTP_CLIENT;
        dport = BOOTP_SERVER;
    }

    /* Create filter. */

    snprintf(filter_buff, sizeof(filter_buff), 
             "arp or icmp or (udp and (src port %d or dst port %d))",
             dport, sport);

    if((dc->rawnet = rawnet_create(interface, filter_buff, sport, dport, promiscuous)) == NULL) {
        destroy_dhcp_client_control(dc);
        return NULL;
    }

    /* Additional fields. */
    
    /* xid setting */
    update_dhcp_xid(dc);
    
    /* secs settings. */
    dhcp_client_reset_secs(dc);
    dhcp_client_update_secs(dc);

    /* discover offer retries. */
    dhcp_client_reset_discover_offer_retries(dc);

#ifdef HAVE_UNAME
    uname(&utsname);
    dc->class_id = splice_many_strings(4, utsname.sysname, utsname.nodename, utsname.release, utsname.version, utsname.machine);
#else /* HAVE_UNAME */
    dc->class_id = strdup("Unknown");
#endif /* HAVE_UNAME */

    if(rawnet_get_hw_addr(dc->rawnet, &interface_addr))  {
        error_message("could not acquire interface address");
        destroy_dhcp_client_control(dc);
        return NULL;
    }

    dc->client_id = create_client_id(interface_addr);
    dc->discover_offer_retries = 0; /* reset discover offer. */
    return dc;
}

/* use fake address in 0x:0x:0x:0x:0x we need to notify rawnet about this.
 * achtung: all packets _except_ dhcp packets will still use the real hardware address. */
void dhcp_control_use_fake_hw_addr(dhcp_client_control_t *dc, char *fake_hw_addr)
{
    char *new_client_id = create_fake_client_id(fake_hw_addr);
    
    if(dc->client_id != NULL)
       xfree(dc->client_id);
    dc->client_id = new_client_id;
    
    rawnet_use_fake_hw_addr(dc->rawnet, fake_hw_addr);
    return;
}

/* destructor. */
void destroy_dhcp_client_control(dhcp_client_control_t *dc)
{
    if(dc->rawnet)
        rawnet_destroy(dc->rawnet);

    if(dc->cache)
        client_cache_destroy(dc->cache);
    
    if(dc->conf)  
        client_conf_destroy(dc->conf);

    if(dc->interface)
        xfree(dc->interface);

    if(dc->class_id)
        xfree(dc->class_id);
    
    if(dc->client_id)
        xfree(dc->client_id);
    
    xfree(dc);
}

/* Utility routines to bring down interface.
 * we need this because we have to be explicit
 * about it. We can't do this at the time we
 * destroy the control object -- for example
 * we destroy the control object through forks
 * to clean out our environment but we don't want
 * to down the interface (necessarily) then.
 * this routine just makes it explicit so we call
 * it when we mean it. */

void dhcp_client_interface_down(dhcp_client_control_t *dc)
{
    rawnet_interface_down(dc->rawnet);
    return;
}
