/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-client-states.c,v 1.31 2002/06/22 01:47:57 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * These are the individual client state functions, 
 * along with their utility routines.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-sysconf.h>
#include <dhcp-globconf.h>
#include <dhcp-net.h>

/* Along with the parameter request list we always build these options
 * upon creating new dhcp packets. */

static list_t *client_build_option_list(unsigned char *opts, unsigned char message_type,
                                        unsigned char *client_id, unsigned char *class_id)
{
    list_t *options = NULL;
    dhcp_option_t *option;
    char *hostname;

    if(opts != NULL)  {
        /* parameter_request_list */
        option  = dhcp_build_parameter_request_list_option(opts);
        options = add_to_list(options, option);
    }
    
    /* dhcp max DHCP message size. */
    /* FIXME: when overloading and the mtu is fixed
     * we also need to fix this magic number. */
    
    option = dhcp_build_max_message_size_option(576);
    options = add_to_list(options, option);
    
    /* dhcp message type. */
    option = dhcp_build_message_type(message_type);
    options = add_to_list(options, option);

    /* client_id */
    option = dhcp_build_client_id(client_id, (ETH_ADDR_LEN + 1));
    options = add_to_list(options, option);

    /* class_id */
    option = dhcp_build_class_id(class_id);
    options = add_to_list(options, option);

    /* hostname if available */
    hostname = glob_conf_get_val(CLIENT_HOSTNAME);
    if(hostname != NULL)   {
        /* if we have one set it. */
        option = dhcp_build_hostname(hostname);
        options = add_to_list(options, option);
    }

    return options;
}

/* Basic DHCP test on incoming rawnet packet. */
static int client_check_dhcp(dhcp_client_control_t *dc)
{
    /* check for generic dhcp response.  
     * we check for type, validity and 
     * matching xid. */

    if((!rawnet_is_valid(dc->rawnet)) ||
       dc->rawnet->type != RAWNET_DHCP ||
       (dhcp_get_xid(dc->rawnet->dhcp_p) != dc->xid))
        return 0;

    /* great: valid dhcp response. */
    return 1;

}

/* check for valid discover responses. */
static int client_check_discover(void *arg)
{
    dhcp_client_control_t *dc = arg;

    if(client_check_dhcp(dc) ||
       dhcp_is_type(dc->rawnet->dhcp_p, DHCP_OFFER_TM) ||
       dhcp_have_atleast_requested_options(dc->rawnet->dhcp_p, dc->conf->options))
        return 1;
    else
        return 0;
}

/* check for valid request responses. */
static int client_check_request(void *arg)
{
    dhcp_client_control_t *dc = arg;

    if(client_check_dhcp(dc) ||
       dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPACK_TM) ||
       dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPNAK_TM))
        return 1;
    else
        return 0;
}

/* 
 * Use this to send out a broadcast reply claiming that we own the
 * IP. To prevent a poisoned cache we use the actual interface hw
 * addr as opposed to any fake one we may have been passed.
 * 
 */

static void client_broadcast_arp_reply(dhcp_client_control_t *dc)
{
    ip_addr_t client_ip_addr;
    eth_addr_t client_hw_addr;

    rawnet_get_ip_addr(dc->rawnet, &client_ip_addr);
    rawnet_get_real_hw_addr(dc->rawnet, &client_hw_addr);

    build_arp_reply_broadcast(dc->rawnet, client_ip_addr, client_hw_addr);
    rawnet_send_packet(dc->rawnet);

    return;
}

/* UNARP (rfc 1868). This makes us more network friendly. */
static void client_broadcast_unarp(dhcp_client_control_t *dc)
{
    ip_addr_t client_ip_addr;
    eth_addr_t client_hw_addr;

    rawnet_get_ip_addr(dc->rawnet, &client_ip_addr);
    rawnet_get_real_hw_addr(dc->rawnet, &client_hw_addr);

    build_unarp(dc->rawnet, client_ip_addr, client_hw_addr);
    rawnet_send_packet(dc->rawnet);
    
    return;
}

/* update secs field. */
static void client_update_packet(void *arg)
{
    dhcp_client_control_t *dc = arg;

    dhcp_client_update_secs(dc);
    rawnet_dhcp_update(dc->rawnet, (uint16_t)dc->secs);
    return;
}

/* do discover/offer transaction. */
int client_discover_offer(dhcp_client_control_t *dc)
{
    list_t *options = NULL;

    options = client_build_option_list(dc->conf->options, DHCP_DISCOVER_TM, 
                                       dc->client_id, dc->class_id);
    
    /* when we begin transmitting discover_offer we setup our secs field.
     * this is because relay agents and dhcp servers use this to see how
     * bad off we are. */

    dhcp_client_reset_secs(dc);
    dhcp_client_update_secs(dc);

    /* Our options cleared by build_* since they are passed
     * down and later purged. */

    build_dhcp_discover(dc->rawnet, dc->xid, dc->secs, options);
    
    if(rawnet_packet_transact(dc->rawnet, dc, client_update_packet, 
                              client_check_discover))  {
        error_message("timeout on DHCP DISCOVER.");
        return STATE_FATAL_ERROR;
    }
    
    /* Check for two options before inserting them ourselves.
     * The server may have decided to be nice to us. */
        
    if(!dhcp_have_option(dc->rawnet->dhcp_p, TAG_DHCP_REQUEST_IP_ADDRESS))
        dc->rawnet->dhcp_p->options = 
        add_to_list(dc->rawnet->dhcp_p->options, 
                    dhcp_build_requested_ip_address(dhcp_get_yiaddr(dc->rawnet->dhcp_p)));
    
    if(!dhcp_have_option(dc->rawnet->dhcp_p, TAG_DHCP_SERVER_IDENTIFIER))
        dc->rawnet->dhcp_p->options = 
        add_to_list(dc->rawnet->dhcp_p->options, 
                    dhcp_build_server_identifier(dhcp_get_siaddr(dc->rawnet->dhcp_p)));

    /* dump cache file. */
    if(client_cache_dump_options(dc->cache, dc->rawnet->dhcp_p->options))  {
        error_message("could not create temporary cache");
        return STATE_FATAL_ERROR;
    }
    
    client_cache_update(dc->cache);
    
    return STATE_REQUEST_ACK;
}

/* do client_request_ack transaction. */
int client_request_ack(dhcp_client_control_t *dc)
{
    list_t *options, *cache_options;

    /* Build options along with the options in our cache. */
    
    options = client_build_option_list(NULL, DHCP_REQUEST_TM, dc->client_id, dc->class_id);
    cache_options = client_cache_load_option_network_list(dc->cache, 0);

    if(cache_options == NULL)  {
        error_message("Could not load cache! This is not good!");
        return STATE_FATAL_ERROR;
    }
    
    options = join_lists(options, cache_options);

    /* Our options cleared by build_* since they are passed
     * down and later purged. */

    build_dhcp_request(dc->rawnet, dc->xid, dc->secs, options);

    if(rawnet_packet_transact(dc->rawnet, dc, NULL, client_check_request))  {
        error_message("timeout on DHCP REQUEST.");
        return STATE_FATAL_ERROR;
    }

    /* if we got a pack we're good to go, if we haven't
     * then we need to revert to discover/offer state. */

    if(dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPACK_TM))
        return STATE_SETUP;
    else
        return STATE_DISCOVER_OFFER;

}

/* do setup (call do_sysconf) */
int client_setup(dhcp_client_control_t *dc)
{
    list_t *options;
    
    options = client_cache_load_option_string_list(dc->cache, 0);
    if(options == NULL)  {
        error_message("received empty configuration cache! cannot configure host");
        return STATE_FATAL_ERROR;
    }

    /* 
     * Before doing a sysconf, let's make sure no one on the
     * local subnet is using our address. Do an ARP request
     * on the local subnet. and check to see if the address
     * is already taken. FIXME: -- this isn't implemented :-) 
     */

    do_sysconf(options, dc);
    cache_entry_purge_list(options);

    client_broadcast_arp_reply(dc);
    
    return STATE_WAIT;
}

/* wait for renewal time. */
int client_wait(dhcp_client_control_t *dc)
{
    struct timeval alarm_time;
    /* 
     * setup signal based on renewal time
     * we actually don't care about lease
     * time since we purge our cache
     * when we exit and release our lease.
     *
     * FIXME: (check RFC compliance)
     */

    rawnet_down(dc->rawnet);


    alarm_time.tv_sec = dc->renewal_time;
    alarm_time.tv_usec = 0;

    set_alarm(alarm_time);

    suspend_for_interrupts(); /* if we got a shutdown or HUP it will be caught later. */

    rawnet_up(dc->rawnet);
    
    /* Update our DHCP secs datum since we're renewing after this. */

    dhcp_client_reset_secs(dc);
    dhcp_client_update_secs(dc);

    return STATE_REQUEST_ACK; /* if we didn't an ALARM get it it's probably a HUP 
                               * or a TERM. in either case we can
                               * return STATE_REQUEST_ACK here. */
}

/* release lease. */
int client_release(dhcp_client_control_t *dc)
{
    list_t *options, *cache_options;
    
    options = client_build_option_list(NULL, DHCP_RELEASE_TM, dc->client_id, dc->class_id);
    cache_options = client_cache_load_option_network_list(dc->cache, 0);
    if(cache_options == NULL)
        /* if we have no cache we probably timed out
         * somewhere. Just return. */
        return STATE_FATAL_ERROR;
    
    options = join_lists(options, cache_options);

    /* Our options cleared by build_* since they are passed
     * down and later purged. */

    build_dhcp_request(dc->rawnet, dc->xid, dc->secs, options);
    
    if(rawnet_send_packet(dc->rawnet) < 0)  {
        error_message("could not send packet: %s", strerror(errno));
        return STATE_FATAL_ERROR;
    }

    client_cache_delete_cache(dc->cache);

    /* We should still have our interface up, so go ahead and send UNARP before
     * bringing it down later. */

    client_broadcast_unarp(dc);

    /* since we're releasing the lease we should down the interface */
    dhcp_client_interface_down(dc);
    
    return 0; /* called from do_shutdown() so no real state returned */
}
