/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-client.c,v 1.33 2002/06/23 22:50:19 bk-rtfm Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-globconf.h>
#include <dhcp-files.h>
#include <dhcp-client.h>

/* forward declarations of do_ routines. */

static void do_version(char *interface);
static void do_kill_client(char *interface);
static void do_wake_client(char *interface);
static void do_clear_cache(char *interface);
static void do_client(char *interface);

/* command to string table. */
const char *command_string[] = {
    "version", "kill",
    "wake", "clear", "client"
};

/* table of functions indexed by command codes. */
client_command commands[] = {
    do_version,        /* print out version information. */
    do_kill_client,    /* kill client. */
    do_wake_client,    /* wake client. */
    do_clear_cache,    /* clear cache. */
    do_client,         /* initialize and run client. */
};

/* forward declaration for interface routines. */
static char *interface_get_active(void);
static char *interface_get_inactive(void);

/* Our get interface sister routine.
   Different command codes will want
   to select interface on different
   criteria. We use this to manage it. */

interface_lister get_interface[] = {
    NULL,                    /* version output needs nothing. */
    interface_get_active,    /* kill client. */
    interface_get_active,    /* wake client. */
    interface_get_active,    /* clear cache. */
    interface_get_inactive,  /* run client. */
};

/* global vars affecting other code. */
int interactive = 1; /* by default we begin interactive (messages on stdout/stderr). */
char *work_dir = CLIENT_WORK_DIR; /* our default working directory */

#if !defined(HAVE_PROGNAME)
const char *__progname;
#endif /* HAVE_PROGNAME */

/* local vars for dhcpclient. */
static char *interface = NULL;            /* the interface we're concerned with. */
static unsigned char want_background = 1; /* whether we should go into the background. */
static char *fake_hw_addr = NULL;         /* if using fake hardware address set here. */
static int promiscuous = 0;               /* whether or not we're running in promiscious mode. */

/* Utility routines. */

/* read up on our global conf. */
static void read_global_conf(const char *interface)
{
    char *filename;

    filename = splice_string(interface, ".conf");
    if(init_glob_conf(filename))
        fatal_error("could not read nor create global configuration file: %s", filename);

    return;
}

/* initialize state per contents of cache. */
static int initialize_client_state(dhcp_client_control_t *dc)
{
    if(client_cache_is_empty(dc->cache))  {
        return STATE_DISCOVER_OFFER;
    } else /* if not empty we assume we can request the options. */
        return STATE_REQUEST_ACK;
}

/* print usage info. */
static void usage(char *s)
{
    printf("usage: %s [-cavkw] [-d directory] [-i interface name] [ -m mac address ]\n", s);
    exit(0);
}

/* print out version information and exit. */
static void do_version(char *interface)
{
    info_message("%s version: %s\n", getprogname(), VERSION);
}

/* do a graceful shutdown. */
static void do_shutdown(dhcp_client_control_t *dc)
{
    info_message("shutting down.");

    if(dc) { /* only do clean up if we can. */
        client_release(dc);
        delete_pid_file(dc->interface);
        destroy_dhcp_client_control(dc);
    }

    exit(0);
}

/* check for existing client. */
static int client_process_exists(char *interface)
{
    pid_t pid;

   if(!get_pid_file(interface, &pid))
       return 0; /* no pid file -- therefore no process. */

#if !defined(KILL_SIGNAL_DETECT)

    /* if we can't detect it assume it is up since the PID file exists :| 
     * issue warning though. */
   warn_message("%s PID file exists please delete and kill process.", getprogname());
   return 0;

#else
       if(!kill(pid, 0))
           return 1;
       else {
           delete_pid_file(interface);
           return 0;
       }
#endif

}

/* wrapper around interface retrieval routines. */

/* get active or inactive interfaces. */
static char *interface_get_proc(int active)
{
    char *first_interface;
    list_t *interface_list;

    if(active)
        interface_list = interface_get_active_interfaces();
    else
        interface_list = interface_get_inactive_interfaces();

    if(interface_list == NULL)
        return NULL;

    first_interface = strdup(interface_list->data);
    purge_list(interface_list, NULL);
    return first_interface;
}

/* get interface which is active. */
static char *interface_get_active(void)
{
    return(interface_get_proc(1));
}

/* get interface which is inactive. */
static char *interface_get_inactive(void)
{
    return(interface_get_proc(0));
}

/* main loop: if only_setup is called we exit after setup --
 * this is useful for running through the loop only long
 * enough to setup the interface. */
static int do_client_dhcp_loop(int only_setup, dhcp_client_control_t *dc, int state)
{
    unsigned char have_setup = 0;
    
    while(1)  {
        
        /* In between states we're clean.
         * 
         * If we get interrupted inside a state
         * we still should be clean, but we don't
         * want to shutdown if we can help it.
         * 
         * Handle any interrupts here. */
        
        check_for_interrupts();

        if(want_shutdown)
            do_shutdown(dc);

        /* if we're only running to be setup, 
         * return here. */

        if(have_setup && only_setup)  {
            return state;
        }
        
        switch(state)  {

            case STATE_FATAL_ERROR:
              info_message("I've encountered a fatal error. I'm giving up.");
              do_shutdown(dc);

           case STATE_DISCOVER_OFFER:
             state = client_discover_offer(dc);
             break;
            
           case STATE_REQUEST_ACK:
             state = client_request_ack(dc);
             break;

           case STATE_SETUP:
             state = client_setup(dc);
             have_setup = 1;
             break;

           case STATE_WAIT:
            state = client_wait(dc);
            break;

           case STATE_RELEASE:
            state = client_release(dc);
            break;

            default:
            break;
        }
    }
    
    return STATE_FATAL_ERROR; /* we should never get here, but if we do... */
}

/* kill a running client. */
static void do_kill_client(char *interface)
{
    pid_t pid;
    
    info_message("killing %s running on interface: %s", getprogname(), interface);

    if(get_pid_file(interface, &pid))  {
        error_message("could not get %s PID.", getprogname());
        error_message("maybe it's not running?");
        return;
    }

    if(kill(pid, SIGTERM) < 0)  {
        error_message("could not send signal to %s process.", getprogname());
        error_message("maybe it's not running?");
        return;
    }

    info_message("killed client on pid: %u", pid);
    return;
}

/* wake up a client. send sigalarm. */
static void do_wake_client(char *interface)
{
    pid_t pid;
    
    if(get_pid_file(interface, &pid))  {
        error_message("could not get %s PID.", getprogname());
        error_message("maybe it's not running?");
        return;
    }
    
    if(kill(pid, SIGALRM) < 0)  {
        error_message("could not send signal to %s process.", getprogname());
        error_message("maybe it's not running?");
        return;
    }

    info_message("woke client on pid %u : for interface: %s", pid, interface);
    return;
}

/* delete existing cache. make sure no other client is running. */
static void do_clear_cache(char *interface)
{
    pid_t pid;
    dhcp_client_control_t *dc;
    client_cache_t *cc;

    if(!get_pid_file(interface, &pid))  {
        /* A race condition exists between the discover and request state.
         * And we can't hup the client and tell it the cache has been deleted. */
        error_message("%s already running.", getprogname());
        error_message("I won't delete cache until it's shutdown.");
        return;
    }
    
    /* As long as no client is running we can go ahead and delete
     * the cache.  We need a client cache structure though, pass
     * it a dummy structure with the interface. That's all it
     * needs. (this a hack).
     */
    
    dc = create_dhcp_client_control_dummy(interface);
    cc = create_client_cache(dc);
    
    /* now delete. */
    client_cache_delete_cache(cc);
    client_cache_delete_tmp_cache(cc);
    
    /* all done. */
    destroy_dhcp_client_control(dc);
    client_cache_destroy(cc);
    
    return;
}

/* change int our working directory. */
static void do_change_work_dir(void)
{
    /* Switch into our work directory */
    
    if(chdir(work_dir) < 0)  {

        /* If it isn't there create it. 
         * Since we're not setuid we
         * assume the root user knows what he's 
         * doing and that it's
         * safe to create this directory.
         * 
         */
        
        if(mkdir(work_dir, S_IRWXU) < 0)
            fatal_error("cannot create work directory", strerror(errno));

        if(chdir(work_dir) < 0)
            fatal_error("cannot enter work directory", work_dir);
    }

    return;
}

/* do client: perform dhcp client proper functions. */
static void do_client(char *interface)
{
    dhcp_client_control_t *dc;
    int state;

    info_message("starting for interface %s", interface);

    /* We need to read our global configuration right after we change to work directory. */
    read_global_conf(interface);

    /* Now see if another client is running on the same interface. */

    if(client_process_exists(interface))  {
        error_message("dhcpclient already exists on interface %s use -k to kill it", interface);
        exit(1);
    }

    /* setup umask: this is also done in go_background but the client
     * doesn't go into the background immediately. so do it now. */

    umask(0);

    /* 
     * Create our control object.
     * There are several things it does
     * for us:
     * 
     * Read the cache, and initialize our state.
     * Read configuration info, and initialize it if needed.
     * Acquire a handle on the interface control.
     * Acquire a handle on the rawnet driver.
     * 
     */

    /* "and she came from a great Desktop; and she razed the
     *  posers leaving a blanket of ash; and we were Praised in
     *  our Judgements;" -- from the Book of Ada (blessed be her
     *  semicolons)  */

    setup_interrupt_handlers();  /* setup signal handling */

    if((dc = create_dhcp_client_control(interface, promiscuous)) == NULL)  {
        error_message("encountered fatal error. cannot continue. exiting");
        do_shutdown(NULL);
    }

    if(fake_hw_addr) /* Marla, you liar, you big tourist, I need this. Now get out! */
        dhcp_control_use_fake_hw_addr(dc, fake_hw_addr);
    
    /* We're good to go. Look up state. */
    state = initialize_client_state(dc);
    
    /* do first loop, returning after setup. */
    state = do_client_dhcp_loop(1, dc, state);

    /* flush stdout to get messages out _now_
     * in case things break in forking into child. */
    fflush(stdout);
    fflush(stderr);

    /* Go in background now before main loop. */
    if(want_background)   { /* are we interactive? if so stay in foreground. 
                             * ... otherwise ... */

        destroy_dhcp_client_control(dc);

        interactive = 0;
        go_background(work_dir);
        setup_interrupt_handlers();  /* setup signal handling,
                                      * we lose them after we fork in go_background. */

        if((dc = create_dhcp_client_control(interface, promiscuous)) == NULL)  {
            error_message("encountered a fatal error. cannot continue. exiting");
            do_shutdown(NULL);
        }

        create_pid_file(interface);
        
        if(fake_hw_addr) /* Marla, you liar, you big tourist, I need this. Now get out! */
            dhcp_control_use_fake_hw_addr(dc, fake_hw_addr);
    }

    do_client_dhcp_loop(0, dc, state); /* perform infinite waiting loop. */
    exit(0);
}

int main(int argc, char *argv[])
{
    int c, command_code = DO_CLIENT;

#if !defined(HAVE_PROGNAME)
    __progname = argv[0];
#endif /* HAVE_PROGNAME */

    info_message("(C) 2001 Thamer Al-Harbash <tmh@whitefang.com>");
    info_message("See LICENSE file for details.");
    info_message(" ");

    while((c = getopt(argc, argv, "pgcdavi:m:kwh:l:H:")) != -1)  {
        
        switch(c)  {
          
          /* check for command codes. */
          case 'v': /* print version. */
            command_code = DO_VERSION;
            break;

          case 'k': /* kill client. */
            command_code = DO_KILL;
            break;

          case 'w': /* wake client. */
            command_code = DO_WAKE;
            break;

          case 'c': /* clear up cache. */
            command_code = DO_CLEAR;
            break;

          case 'd':
            work_dir = strdup(optarg);
            break;

          /* setup options. */

          case 'a':
            want_background = 0;
            break;
            
          case 'i':
            interface = strdup(optarg);
            break;
            
          case 'm':
            fake_hw_addr = strdup(optarg);
            break;

          case 'l':

            if(set_verbosity_level(atoi(optarg)))  {
                error_message("illegal verbosity level: %s", optarg);
                exit(1);
            }
            
            break;

          case 'p':
            promiscuous = 1;
            break;

          /* dhcp options which can be set on the command line. */
          case 'H':
            glob_conf_set_val(optarg, CLIENT_HOSTNAME);
            break;

          case 'h': /* fall through. */
          default:
            usage(argv[0]);
            break;
        }
    }

    do_change_work_dir();


    /* get interface if not specified. */
    if(interface == NULL)  {
        if(get_interface[command_code] != NULL) 
            interface = get_interface[command_code]();
    }

    if(interface == NULL && command_code != DO_VERSION)   {
        /* if we were unable to attain an interface. */
        error_message("unable to retrieve suitable interface");
        error_message("for this operation: %s", command_string[command_code]);
        exit(1); /* can't use fatal_error for multiline error message. */
        /* fixme: this stupidity will go with exception stack. */
    }

    commands[command_code](interface);
    exit(0);
}
