/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-com.c,v 1.9 2002/06/17 13:23:07 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * dhcp object.
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>

static const unsigned char dhcp_magic_cookie[4] = { 99, 130, 83, 99 };
    
/* create new dhcp object. */

dhcp_obj *dhcp_create(void)
{
    dhcp_obj *dhcp;
    
    dhcp = xmalloc(sizeof(dhcp_obj));
    dhcp->options = NULL;
    dhcp->options_seek = NULL;

    return dhcp;
}

/* clean out options. */

void dhcp_purge_option_list(list_t *options)
{
    list_t *list_ptr, *tmp;
    dhcp_option_t *option;

    /* Free up options. */
    
    if(options)  {
        
        list_ptr = options;
        
        while(list_ptr)  {
            
            tmp = list_ptr->next;
            option = list_ptr->data;

            xfree(option->data);
            xfree(option);
            xfree(list_ptr);

            list_ptr = tmp;
        }
    }
    
    return;
}

/* Clean out dhcp object so it can be reused again. */

void dhcp_purge(dhcp_obj *dhcp)
{
    dhcp_purge_option_list(dhcp->options);
    dhcp->options = NULL;
    dhcp->options_seek = NULL;
    
    return;
}

/* Purge and free up dhcp object. */

void dhcp_destroy(dhcp_obj *dhcp)
{
    dhcp_purge(dhcp);
    xfree(dhcp);
    return;
}

/* create option. */
dhcp_option_t *create_dhcp_option(const unsigned char *opt_data, unsigned char len, unsigned char tag)
{
    dhcp_option_t *option;
    
    option = xmalloc(sizeof(dhcp_option_t));
    option->tag = tag;
    option->len = len;
    
    option->data = xmalloc(option->len);
    memcpy(option->data, opt_data, option->len);

    return option;
}

/* destroy dhcp option. */
void destroy_dhcp_option(dhcp_option_t *option)
{
    xfree(option->data);
    xfree(option);
    
    return;
}

/* copy DHCP option safely. */
static dhcp_option_t *process_next_dhcp_option(const unsigned char *opt_data,
                                             int len, unsigned char *optlen)
{
    dhcp_option_t *option;

    /* Check for malformed option. */
    if(len < 2 || !dhcp_option_is_valid(opt_data[0], opt_data[1]))
        return NULL;

    /* We're in the clear copy out the bytes into a new
     * dhcp option object. */

    option = create_dhcp_option(&opt_data[2], opt_data[1], opt_data[0]);
    *(optlen) = option->len;
    return option;
}

static void dhcp_read_options_image(dhcp_obj *dhcp, const unsigned char *dhcp_packet, int len)
{
    unsigned char optlen;
    dhcp_option_t *option;

    /* Try and copy all the options.
     * We do no parsing at this point
     * since we may not require all
     * the options (if any).
     * 
     * From here on in we cannot return an error
     * even if the options are malformed.
     * We'll just do our best.
     * 
     */

    while(len > 0)  {
        
        /* Always skip pads */ 
        
        if(*dhcp_packet == TAG_PAD)  {
            dhcp_packet++;
            len--;
            continue;
        }
        
        /* Always stop on end. */

        if(*dhcp_packet == TAG_END)
            break;

        /* If not PAD or END we copy the data out. */

        option = process_next_dhcp_option(dhcp_packet, len, &optlen);
        if(option == NULL)
            break;

        dhcp->options = add_to_end_of_list(dhcp->options, option);

        len -= (optlen + 2);
        dhcp_packet += (optlen + 2);
        
    }
    
    dhcp->options_seek = dhcp->options;
    
    return;
}

/* read packet image. */
int dhcp_read_packet_image(dhcp_obj *dhcp, const unsigned char *dhcp_packet, int len)
{
    /* Get the fixedheader or if not
     * possible return an error. */
    
    if(len < DHCP_FIXEDHDR_LEN)
        return -1;
    else
        align_dhcphdr(dhcp_packet, &dhcp->fixedheader);

    len -= DHCP_FIXEDHDR_LEN;
    dhcp_packet += DHCP_FIXEDHDR_LEN;

    /* Copy out magic cookie. 4 octets. */

    if(len >= 4)
        memcpy(&dhcp->magic_cookie, dhcp_packet, 4);
    else  {
        memset(&dhcp->magic_cookie, 0, 4); /* set it to 0 so it doesn't pass test. */
        return 0;
    }
    
    len -= 4;
    dhcp_packet += 4;
    
    /* read options placed after cookie. */
    dhcp_read_options_image(dhcp, dhcp_packet, len);

    /* We need to check file before sname to be compliant. */

    /* Now check if we have options overload in file field
     * and read options as necesssary. */
    if(dhcp_is_file_overload(dhcp))
        dhcp_read_options_image(dhcp, dhcp->fixedheader.file, DHCP_FILE_SIZE);

    /* Ditto on the sname field. */
    if(dhcp_is_sname_overload(dhcp))
        dhcp_read_options_image(dhcp, dhcp->fixedheader.sname, DHCP_SNAME_SIZE);

    return 0;
}

/* Accessor methods.
 * 
 * Do conversion for some of the datatypes.
 * 
 * The exceptions are ipv4 addresses.
 * We leave those in network order.
 */

/* dhcp op */
unsigned char dhcp_get_op(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.op);
}
/* dhcp hardware type */
unsigned char dhcp_get_htype(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.htype);
}

/* dhcp hlen type */
unsigned char dhcp_get_hlen(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.hlen);
}

/* dhcp hops */
unsigned char dhcp_get_hops(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.hops);
}

/* dhcp xid: do conversion */
uint32_t dhcp_get_xid(dhcp_obj *dhcp)
{
     return(ntohl(dhcp->fixedheader.xid));
}

/* dhcp secs: do conversion. */
uint16_t dhcp_get_secs(dhcp_obj *dhcp)
{
     return(ntohs(dhcp->fixedheader.secs));
}

/* dhcp flags: do conversion */
uint16_t dhcp_get_flags(dhcp_obj *dhcp)
{
     return(ntohs(dhcp->fixedheader.flags));
}

/* dhcp ciaddr: no conversion necessary */
uint32_t dhcp_get_ciaddr(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.ciaddr);
}

/* dhcp yiaddr: no conversion necessary */
uint32_t dhcp_get_yiaddr(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.yiaddr);
}

/* dhcp siaddr: no conversion necessary */
uint32_t dhcp_get_siaddr(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.siaddr);
}

/* dhcp ciaddr: no conversion necessary */
uint32_t dhcp_get_giaddr(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.giaddr);
}

/* dhcp chaddr */
unsigned char *dhcp_get_chaddr(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.chaddr);
}

/* dhcp giaddr */
unsigned char *dhcp_get_sname(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.sname);
}

/* dhcp file */
unsigned char *dhcp_get_filename(dhcp_obj *dhcp)
{
     return(dhcp->fixedheader.file);
}
/* return magic cookie */
uint32_t dhcp_get_magic_cookie(dhcp_obj *dhcp)
{
    return(dhcp->magic_cookie);
}

/* Options routines. */

/* Reset the options seek point. */

void dhcp_reset_option_seek(dhcp_obj *dhcp)
{
    dhcp->options_seek = dhcp->options;
}

/* Return the next option or NULL if at the end. */

dhcp_option_t *dhcp_get_next_option(dhcp_obj *dhcp)
{
    dhcp_option_t *rval;

    if(dhcp->options_seek == NULL)
        return NULL;
    
    rval = dhcp->options_seek->data;
    
    dhcp->options_seek = dhcp->options_seek->next;
    
    return rval;
}

/* tests. */

/* test for valid cookie. */
int dhcp_valid_magic_cookie(dhcp_obj *dhcp)
{

    if(!memcmp(&dhcp->magic_cookie, dhcp_magic_cookie, 4))
        return 1;
    else
        return 0;

}

/* test for type. */
int dhcp_is_type(dhcp_obj *dhcp, unsigned char type)
{
    dhcp_option_t *opt;
    
    dhcp_reset_option_seek(dhcp);

    while((opt = dhcp_get_next_option(dhcp)) != NULL)  {

        if(opt->tag == TAG_DHCP_MESSAGE_TYPE)  {
            if(opt->len != 1) /* let's not get hosed shall we? */
                continue;

            if(*(opt->data) == type)
                return 1;
            else 
                break;
        }
    }

    return 0;
}

/* test for option validity. */
int dhcp_option_is_valid(unsigned char tag, unsigned char tag_len)
{
    if(tag_len < 1 || tag >= MAX_OPTIONS_HANDLED)
        return 0;
    else
        return 1;
}

/* test for existance of option. */
int dhcp_have_option(dhcp_obj *dhcp, unsigned char tag)
{
    dhcp_option_t *option;
    
    /* Check for a option by tag. */
    dhcp_reset_option_seek(dhcp);
    
    while((option = dhcp_get_next_option(dhcp)) != NULL)  {
        if(option->tag == tag)
            return 1;
    }
    
    return 0;
}

/* get the overload which may be for file, sname, or both. */
static unsigned char dhcp_get_overload_option(dhcp_obj *dhcp)
{
    dhcp_option_t *option;
    
    dhcp_reset_option_seek(dhcp);
    
    while((option = dhcp_get_next_option(dhcp)) != NULL)  {
        if(option->tag == TAG_DHCP_OPTION_OVERLOAD)  {
            return *(option->data);
        }
    }

    return 0;
}

/* options overloaded on file field? */
int dhcp_is_file_overload(dhcp_obj *dhcp)
{
    unsigned char overload_val = dhcp_get_overload_option(dhcp);
    
    if(overload_val == DHCP_OVERLOAD_FILE ||
       overload_val == DHCP_OVERLOAD_BOTH)
        return 1;
    else
        return 0;
}

/* options overloaded on sname field? */
int dhcp_is_sname_overload(dhcp_obj *dhcp)
{
    unsigned char overload_val = dhcp_get_overload_option(dhcp);

    if(overload_val == DHCP_OVERLOAD_SNAME ||
       overload_val == DHCP_OVERLOAD_BOTH)
        return 1;
    else
        return 0;
}

    
/* Check if we have at least the requested options:
 * count up the number of options we want, then
 * check for matches. If we match as many as we
 * have configured then we're ok. */

int dhcp_have_atleast_requested_options(dhcp_obj *dhcp, unsigned char *options)
{
    int i;

    for(i = 0;i < MAX_OPTIONS_HANDLED;i++)  {
        if(options[i] && !dhcp_have_option(dhcp, i))
                return 1;
    }
    
    return 0;
}
        
/* This is a little tricky since we haven't written anything
 * yet. We know that we must end on a 32-bit boundary, and
 * we know that we must include the tag, len on each option
 * as well as the end option '255' */

uint16_t dhcp_get_options_len(list_t *options)
{
    uint16_t len = 0;
    list_t *opts;
    dhcp_option_t *opt;
    int padding;

    for(opts = options;opts;opts = opts->next)  {
        opt = opts->data;

        len += 2; /* for tag and len bytes. */
        len += opt->len;
    }
    
    len += 1; /* For the end option. */

    if((padding = len%4))
        len += padding;

    return len;
}

/* accept bit array telling us which options to build the request option. */
dhcp_option_t *dhcp_build_parameter_request_list_option(unsigned char *requested_options)
{
    unsigned char i, j;
    unsigned char len = 0;
    dhcp_option_t *option;
    
    /* Grab length so we can just malloc all at once. */
    for(i = 0;i < MAX_OPTIONS_HANDLED;i++)  {
        if(requested_options[i])
            len++;
    }
    
    if(len == 0)  {
        error_message("error building parameter request: no options!");
        return NULL;
    }
    
    option = xmalloc(sizeof(dhcp_option_t));
    option->len = len;
    option->tag = TAG_PARAMETERS;
    option->data = xmalloc(len);
    
    j = 0;
    for(i = 0;i < MAX_OPTIONS_HANDLED;i++)  {
        if(requested_options[i])  {
            option->data[j] = i;
            j++;
        }
    }                                     
    
    return option;
}

dhcp_option_t *dhcp_build_max_message_size_option(uint16_t max)
{
    max = htons(max);
    return(create_dhcp_option((unsigned char *)&max, sizeof(uint16_t), TAG_MAX_DHCP_SIZE));
}

dhcp_option_t *dhcp_build_message_type(unsigned char type)
{
    return(create_dhcp_option((unsigned char *)&type, sizeof(unsigned char), TAG_DHCP_MESSAGE_TYPE));
}

dhcp_option_t *dhcp_build_class_id(unsigned char *class_id)
{
    unsigned char len = strlen(class_id);
    return(create_dhcp_option(class_id, len, TAG_DHCP_CLASS_ID));
}

dhcp_option_t *dhcp_build_client_id(unsigned char *client_id, unsigned char len)
{
    return(create_dhcp_option(client_id, len, TAG_DHCP_CLIENT_ID));
}

dhcp_option_t *dhcp_build_requested_ip_address(uint32_t address)
{
    return(create_dhcp_option((unsigned char *)&address, IP_ADDR_LEN, TAG_DHCP_REQUEST_IP_ADDRESS));
}

dhcp_option_t *dhcp_build_server_identifier(uint32_t address)
{
    return(create_dhcp_option((unsigned char *)&address, IP_ADDR_LEN, TAG_DHCP_SERVER_IDENTIFIER));
}

dhcp_option_t *dhcp_build_hostname(char *hostname)
{
    return(create_dhcp_option((unsigned char *)hostname, (strlen(hostname) + 1), TAG_DHCP_HOSTNAME));
}

/* DHCP set methods. */

/* dhcp op */
void dhcp_set_op(dhcp_obj *dhcp, unsigned char op)
{
    memcpy(&dhcp->fixedheader.op, &op, 1);
}

/* dhcp hardware type */
void dhcp_set_htype(dhcp_obj *dhcp, unsigned char htype)
{
    memcpy(&dhcp->fixedheader.htype, &htype, 1);
}

/* dhcp hlen type */
void dhcp_set_hlen(dhcp_obj *dhcp, unsigned char hlen)
{
    memcpy(&dhcp->fixedheader.hlen, &hlen, 1);
}

/* dhcp hops */
void dhcp_set_hops(dhcp_obj *dhcp, unsigned char hops)
{
    memcpy(&dhcp->fixedheader.hops, &hops, 1);
}

/* dhcp xid: do conversion */
void dhcp_set_xid(dhcp_obj *dhcp, uint32_t xid)
{
    xid = htonl(xid);
    memcpy(&dhcp->fixedheader.xid, &xid, 4);
}

/* dhcp secs: do conversion. */
void dhcp_set_secs(dhcp_obj *dhcp, uint16_t secs)
{
    secs = ntohs(secs);
    memcpy(&dhcp->fixedheader.secs, &secs, 2);
}

/* dhcp flags: do conversion */
void dhcp_set_flag_broadcast(dhcp_obj *dhcp)
{
    dhcp->fixedheader.flags = htons(DHCP_BROADCAST_FLAG);
    return;
}

/* dhcp ciaddr: no conversion necessary */
void dhcp_set_ciaddr(dhcp_obj *dhcp, uint32_t ciaddr)
{
    memcpy(&dhcp->fixedheader.ciaddr, &ciaddr, 4);
    return;
}

/* dhcp yiaddr: no conversion necessary */
void dhcp_set_yiaddr(dhcp_obj *dhcp, uint32_t yiaddr)
{
    memcpy(&dhcp->fixedheader.yiaddr, &yiaddr, 4);
    return;
}

/* dhcp siaddr: no conversion necessary */
void dhcp_set_siaddr(dhcp_obj *dhcp, uint32_t siaddr)
{
    memcpy(&dhcp->fixedheader.siaddr, &siaddr, 4);
    return;
}

/* dhcp giaddr: no conversion necessary */
void dhcp_set_giaddr(dhcp_obj *dhcp, uint32_t giaddr)
{
    memcpy(&dhcp->fixedheader.giaddr, &giaddr, 4);
    return;
}

/* dhcp chaddr */
void dhcp_set_chaddr(dhcp_obj *dhcp, unsigned char *chaddr, int len)
{
        memcpy(&dhcp->fixedheader.chaddr, chaddr, len);
}

/* dhcp chaddr */
void dhcp_clear_chaddr(dhcp_obj *dhcp)
{
    memset(dhcp->fixedheader.chaddr, 0, DHCP_CHADDR_SIZE);
    return;
}

/* dhcp sname */
void dhcp_set_sname(dhcp_obj *dhcp, unsigned char *sname)
{
    memcpy(&dhcp->fixedheader.sname, sname, DHCP_SNAME_SIZE);
    return;
}

/* dhcp file */
void dhcp_set_filename(dhcp_obj *dhcp, unsigned char *file)
{
    memcpy(&dhcp->fixedheader.file, file, DHCP_FILE_SIZE);
    return;
}

/* dhcp sname */
void dhcp_clear_sname(dhcp_obj *dhcp)
{
    memset(&dhcp->fixedheader.sname, 0, DHCP_SNAME_SIZE);
    return;
}

/* dhcp file */
void dhcp_clear_filename(dhcp_obj *dhcp)
{
    memset(&dhcp->fixedheader.file, 0, DHCP_FILE_SIZE);
    return;
}

/* cookie. */
void dhcp_set_magic_cookie(dhcp_obj *dhcp)
{
    memcpy(&dhcp->magic_cookie, dhcp_magic_cookie, 4);
    return;
}

/* dhcp set options, add options. */

/* replace options with list passed (convenience). */
void dhcp_set_options(dhcp_obj *dhcp, list_t *options)
{
    dhcp->options = options;
    dhcp->options_seek = options;
    return;
}

/* write options to packet pointer. */
void dhcp_write_options(dhcp_obj *dhcp, unsigned char *packet)
{
    dhcp_option_t *option;
    int padding, len = 0;
    
    dhcp_reset_option_seek(dhcp);
    
    while((option = dhcp_get_next_option(dhcp)) != NULL)  {

        *packet = option->tag;
        packet += 1;
        len++;
        
        *packet = option->len;
        packet += 1;
        len++;
        
        memcpy(packet, option->data, option->len);
        packet += option->len;
        len += option->len;
        
    }

    *packet = TAG_END;
    packet += 1;
    len += 1;

    /* Make sure we pad up. */

    if((padding = len%4))  {
        while(padding--)  {
            *packet = TAG_PAD;
            packet++;
        }
    }

    return;
}

/* write dhcp image out to packet. */
void dhcp_write_packet_image(dhcp_obj *dhcp, unsigned char *packet)
{
    /* Dump fixed header first. */
    
    memcpy(packet, &dhcp->fixedheader.op, 1);
    packet += sizeof(dhcp->fixedheader.op);
    
    memcpy(packet, &dhcp->fixedheader.htype, 1);
    packet += sizeof(dhcp->fixedheader.htype);
    
    memcpy(packet, &dhcp->fixedheader.hlen, 1);
    packet += sizeof(dhcp->fixedheader.hlen);
    
    memcpy(packet, &dhcp->fixedheader.hops, 1);
    packet += sizeof(dhcp->fixedheader.hops);
    
    memcpy(packet, &dhcp->fixedheader.xid, 4);
    packet += sizeof(dhcp->fixedheader.xid);
    
    memcpy(packet, &dhcp->fixedheader.secs, 2);
    packet += sizeof(dhcp->fixedheader.secs);
    
    memcpy(packet, &dhcp->fixedheader.flags, 2);
    packet += sizeof(dhcp->fixedheader.flags);

    memcpy(packet, &dhcp->fixedheader.ciaddr, IP_ADDR_LEN);
    packet += IP_ADDR_LEN;
    
    memcpy(packet, &dhcp->fixedheader.yiaddr, IP_ADDR_LEN);
    packet += IP_ADDR_LEN;
    
    memcpy(packet, &dhcp->fixedheader.siaddr, IP_ADDR_LEN);
    packet += IP_ADDR_LEN;

    memcpy(packet, &dhcp->fixedheader.giaddr, IP_ADDR_LEN);
    packet += IP_ADDR_LEN;

    memcpy(packet, &dhcp->fixedheader.chaddr, DHCP_CHADDR_SIZE);
    packet += DHCP_CHADDR_SIZE;
    
    memcpy(packet, &dhcp->fixedheader.sname, DHCP_SNAME_SIZE);
    packet += DHCP_SNAME_SIZE;
    
    memcpy(packet, &dhcp->fixedheader.file, DHCP_FILE_SIZE);
    packet += DHCP_FILE_SIZE;

    /* Now dump cookie. */
    
    memcpy(packet, &dhcp->magic_cookie, sizeof(dhcp->magic_cookie));
    packet += sizeof(dhcp->magic_cookie);
    
    /* Now write options. */

    dhcp_write_options(dhcp, packet);
    
    return;
}

/* generate xid. */
uint32_t dhcp_gen_xid(void)
{
    return(get_random_uint32());
}
