/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-convert.c,v 1.15 2002/06/17 13:23:07 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* 
 * Convert options to strings for serialization,
 * or from serialized state to strings, or to
 * raw data state.
 * 
 * All functions return freshly allocated datums.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-convert.h>

/* We have four handlers:
 * 
 * Network to serialized: take a datum in raw network format and
 * serialize into serial datum.
 * 
 * Serial to internal: take a serial and convert it to internal
 * format. 
 * 
 * Serial to network: take a serial datum and turn it into a raw
 * network datum.
 * 
 * Free internal: special freeing routine used for internal datum.
 * This is used so we can pass around internal datums elsewhere without
 * having to understand them.
 *
 * All network to serialized handlers do validity checks for length
 * matching exact sizes or multiples of sizes. This is where we expect
 * malformed options to be stomped on.
 * 
 */

option_convert_handler option_convert_handlers[] = {
    { NULL, NULL, NULL, NULL },                                  /* 0 */ /* pad */
    { network_addr_to_string, string_to_addr, string_addr_to_network, 
            xfree },                                             /* 1 */ /* subnet mask */
    { network_int32_to_string, string_to_int32, 
            string_int32_to_network, xfree },                    /* 2 */ /* time offset */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 3 */ /* router. */
    { network_addr_list_to_string, string_to_addr_list,
             string_addr_list_to_network, purge_list_internal }, /* 4 */ /* time server. */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 5 */ /* name server. (old) */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 6 */ /* domain name servers. */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 7 */ /* mit-lcs log server */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 8 */ /* cookie server. */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 9 */ /* LPR server. */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 10 */ /* Impress server. */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 11 */ /* Resource Location Server. */
    { validate_network_string_to_string, 
            strdup_wrap, strdup_wrap_net, xfree },               /* 12 */ /* Host name option. */
    { network_uint16_to_string, string_to_uint16,
            string_uint16_to_network, xfree },                   /* 13 */ /* boot file size option */
    { validate_network_string_to_string, 
            strdup_wrap, strdup_wrap_net, xfree },               /* 14 */ /* merit dump file */
    { validate_network_string_to_string, strdup_wrap, strdup_wrap_net, 
            xfree },                                             /* 15 */ /* domain name */
    { network_addr_to_string, string_to_addr, string_addr_to_network, 
            xfree },                                             /* 16 */ /* swap server */
    { validate_network_string_to_string, strdup_wrap, strdup_wrap_net, 
            xfree },                                             /* 17 */ /* root path */
    { validate_network_string_to_string, strdup_wrap, strdup_wrap_net, 
            xfree },                                             /* 18 */ /* extension path */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 19 */ /* IP forwarding. */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 20 */ /* Non Local Source Routing. */
    { network_addr_pair_list_to_string, string_to_addr_pair_list,
      string_addr_pair_list_to_network, purge_list_internal },   /* 21 */ /* Policy filter option. */
    { network_uint16_to_string, string_to_uint16,
            string_uint16_to_network, xfree },                   /* 22 */ /* maximum dgram reassembly */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 23 */ /* IP TTL */
    { network_uint32_to_string, string_to_uint32, 
            string_uint32_to_network, xfree },                   /* 24 */ /* MTU Aging Timeout. */
    { network_uint16_list_to_string, string_to_uint16_list,
      string_uint16_list_to_network, purge_list_internal },      /* 25 */ /* MTU Plateau Table. */
    { network_uint16_to_string, string_to_uint16,
      string_uint16_to_network, xfree },                         /* 26 */ /* Interface MTU. */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 27 */ /* Subnets are Local */
    { network_addr_to_string, string_to_addr, string_addr_to_network, 
            xfree },                                             /* 28 */ /* Broadcast Address */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 29 */ /* Perform Mask Discovery */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 30 */ /* Mask Supplier */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 31 */ /* Perform Router Discovery */
    { network_addr_to_string, string_to_addr, 
            string_addr_to_network, xfree },                     /* 32 */ /* Router Solicitation Address */
    { network_addr_pair_list_to_string, string_to_addr_pair_list,
      string_addr_pair_list_to_network, purge_list_internal },   /* 33 */ /* Static Route */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 34 */ /* Trailer Encapsulation */
    { network_uint32_to_string, string_to_uint32, 
            string_uint32_to_network, xfree },                   /* 35 */ /* ARP Cache Timeout. */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 36 */ /* Ethernet Encapsulation */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 37 */ /* TCP Default TLL */
    { network_uint32_to_string, string_to_uint32, 
            string_uint32_to_network, xfree },                   /* 38 */ /* TCP Keepalive Interval */
    { network_byte_to_string_byte, string_byte_to_byte, 
            string_byte_to_network, xfree },                     /* 39 */ /* TCP Keepalive Garbage Option. */
    { validate_network_string_to_string, strdup_wrap,
        strdup_wrap_net, xfree },                                /* 40 */ /* Network Information Service Domain */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 41 */ /* Network Information Servers */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 42 */ /* Network Time Protocol Servers. */
    { NULL, NULL, NULL, NULL },                                  /* 43 */ /* Vendor Specific (don't handle yet.) */ 
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 44 */ /* NetBIOS over TCP/IP Name Server */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 45 */ /* NetBIOS over TCP/IP Datagram Dist Server */
    { network_byte_to_string_byte, string_byte_to_byte,
            string_byte_to_network, xfree },                     /* 46 */ /* NetBIOS over TCP/IP Node Type. */
    { validate_network_string_to_string,
            strdup_wrap, strdup_wrap_net, xfree },               /* 47 */ /* NetBIOS over TCP/IP Scope */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 48 */ /* X Windows System Font Server. */
    { network_addr_list_to_string, string_to_addr_list, 
            string_addr_list_to_network, purge_list_internal },  /* 49 */ /* X Window System Display Manager. */
    { network_addr_to_string, string_to_addr, 
            string_addr_to_network, xfree },                     /* 50 */ /* Request IP Address. */
    { network_uint32_to_string, string_to_uint32,
            string_uint32_to_network, xfree },                   /* 51 */ /* IP Address Lease Time */
    { network_byte_to_string_byte, string_byte_to_byte,
            string_byte_to_network, xfree },                     /* 52 */ /* Option Overload. */
    { NULL, NULL, NULL, NULL },                                  /* 53 */ /* Dhcp Message Type. */
    { network_addr_to_string, string_to_addr,
            string_addr_to_network, xfree },                     /* 54 */ /* Server Identifier. */
    { NULL, NULL, NULL, NULL },                                  /* 55 */ /* parameter request list. */
    { NULL, NULL, NULL, NULL },                                  /* 56 */ /* message. we handle this higher. */
    { network_uint16_list_to_string, string_to_uint16_list,
      string_uint16_list_to_network, purge_list_internal },      /* 57 */ /* Maximum DHCP message size. */
    { network_uint32_to_string, string_to_uint32, 
            string_uint32_to_network, xfree },                   /* 58 */ /* Renewal time. */
    { network_uint32_to_string, string_to_uint32, 
            string_uint32_to_network, xfree },                   /* 59 */ /* Rebinding time. */
    { NULL, NULL, NULL, NULL },                                  /* 60 */ /* Server identifier */
    { NULL, NULL, NULL, NULL },                                  /* 61 */ /* Client identifier */
};

/* Byte conversions useful for flags, and 1/0 values. */

/* network byte value to string byte value. */
char *network_byte_to_string_byte(const unsigned char *data, int len)
{
    char *s;
    
    if(len < 1)
        return NULL;
    
    s = xmalloc(BYTE_STRING_SIZE);
    snprintf(s, BYTE_STRING_SIZE, "0x%hhx", *(unsigned char *)data);
    
    return s;
}

/* string to internal byte value. */
void *string_byte_to_byte(const char *s)
{
    unsigned char *b;
    
    b = xmalloc(sizeof(unsigned char));
    
    sscanf(s,"0x%hhx",b);
    
    return b;
}

/* string to network byte value. */
void *string_byte_to_network(const char *s, unsigned char *ret_len)
{
    *ret_len = 1;
    return(string_byte_to_byte);
}

/* IPv4 address conversion routines. */

/* convert one network address datum to serialized form. */
char *network_addr_to_string(const unsigned char *data, int len)
{
    struct in_addr address;
    
    if(len < IP_ADDR_LEN)
        return NULL;
    
    memcpy(&address.s_addr, data, IP_ADDR_LEN);

    return(strdup(inet_ntoa(address)));
}

/* convert list of network address datums to serialized form. */
char *network_addr_list_to_string(const unsigned char *data, int len)
{
    char *string;
    char *tmp, *new_string;
    
    if(len < IP_ADDR_LEN)
        return NULL;
    
    string = network_addr_to_string(data, len);
    if(string == NULL)
        return NULL;
    
    tmp = string;
    string = splice_string(string, ";");
    xfree(tmp);
    
    len -= 4;
    data += 4;
    
    while(len >= 4)  {
        

        tmp = string;
        new_string = network_addr_to_string(data, len);
        
        if(new_string == NULL)  {
            
            /* truncated address. Just do our best. */
            
            string = splice_string(string, ";");
            free(tmp);
            return string;
        }

        string = splice_string(string, new_string);
        xfree(tmp);
        xfree(new_string);

        tmp = string;
        string = splice_string(string, ";");
        xfree(tmp);

        data += 4;
        len -= 4;
    }
    
    return string;
}

/* convert string address to internal address datum. */
void *string_to_addr(const char *string)
{
    uint32_t addr, *new_addr;

    addr = inet_addr(string);
    
    new_addr = xmalloc(IP_ADDR_LEN);
    memcpy(new_addr, &addr, IP_ADDR_LEN);

    return new_addr;
}

/* convert list of string addresses to internal address datum. */
void *string_to_addr_list(const char *string)
{
    uint32_t *addr;
    list_t *addr_list = NULL;
    char *cp, *tmp_string, *orig_string;

    tmp_string = strdup(string); /* don't overwrite old string. */
    trim_string(tmp_string);
    orig_string = tmp_string;
    
    while((cp = strchr(tmp_string, ';')) != NULL)  {

        *cp = 0;
        addr = string_to_addr(tmp_string);
        addr_list = add_to_end_of_list(addr_list, addr);
        cp++;
        
        if(*cp == 0)
            break;

        tmp_string = cp;
    }

    xfree(orig_string);
    return addr_list;
}

/* convert serialized address to network datum. */
void *string_addr_to_network(const char *string, unsigned char *len_ret)
{
    uint32_t *addr = string_to_addr(string);
    *len_ret = 4;
    
    return addr;
}

/* convert serialized address list to network datum. */
void *string_addr_list_to_network(const char *string, unsigned char *len_ret)
{
    list_t *ptr;
    list_t *addr_list = string_to_addr_list(string);
    int count = 0;
    uint32_t *data, *cp;
    
    for(ptr = addr_list;ptr;ptr = ptr->next)
        count++;
    
    data = xmalloc(count * IP_ADDR_LEN);
    cp = data;
    
    for(ptr = addr_list;ptr;ptr = ptr->next)  {
        memcpy(cp, ptr->data, IP_ADDR_LEN);
        cp++;
    }
    
    purge_list(addr_list, NULL);
    
    *len_ret = (count * 4);
    return data;
}

/* address pair conversion. this just adds a check to make sure len is divisible
 * by 2 addresses. */
char *network_addr_pair_list_to_string(const unsigned char *data, int len)
{
    if(len == 0 ||
       len%(IP_ADDR_LEN * 2))
        return NULL;
    
    return(network_addr_list_to_string(data, len));
}

/* string to internal address pair list (wrapper). */
void *string_to_addr_pair_list(const char *s)
{
    return(string_to_addr_list(s));
}

/* string to network address pair list (wrapper). */
void *string_addr_pair_list_to_network(const char *s, unsigned char *retlen)
{
    return(string_addr_list_to_network(s, retlen));
}

/* String conversions:
 * We use strdup mostly to just make copies in our conversion hook
 * array. However, for network strings we want to run them through 
 * validate_network_to_string
 */

/* validate string only: fixme: add rfc character set check. */
char *validate_network_string_to_string(const unsigned char *data, int len)
{
    char *new_string;
    
    if(!is_string(data, len))  { /* fix if needed. */
        new_string = xmalloc(len + 1);
        memcpy(new_string, data, len);
        new_string[len] = 0;
        return new_string;
    }
    
    /* otherwise strdup a copy. */
    return strdup(data);
}

/* validate internal strings -- e.g. conf and cache file. */
char *validate_string_to_string(const unsigned char *data, int len)
{
    /* we don't validate or do much with internal strings.
     * however we need to maintain the prototype to plug this
     * into the right places. */

    return strdup(data);
}

/* just copy. wrapper needed to be nice to the compiler. */
void *strdup_wrap(const char *s)
{
    return(strdup(s));
}

/* call strdup_wrap but set the ret_len appropriately. */
void *strdup_wrap_net(const char *s, unsigned char *ret_len)
{
    unsigned char *val;
    
    val = strdup_wrap(s);

    *ret_len = (strlen(s) + 1);
    return val;
}

/* Integer conversions:
 * More work is done here to convert to network or internal format. */

char *int_to_string(const unsigned char *data, int data_len,
                    size_t size, int si)
{
    char *string;
    
    if(size > data_len)
        return NULL;
    
    if(si)  { /* signed. */
        
        switch(size)  {

          case sizeof(int32_t):
              string = xmalloc(INT32_T_STRING_SIZE);
              snprintf(string, INT32_T_STRING_SIZE, "%"PRIi32"", *(int32_t *)data);
              break;

          case sizeof(int16_t):
              string = xmalloc(INT16_T_STRING_SIZE);
              snprintf(string, INT16_T_STRING_SIZE, "%"PRIi16"", *(int16_t *)data);
              break;

          default:
             /* we should never get here but we should force ourselves
              * to b0rk if we do. */
             fatal_error("illegal size passed to int_to_string()!");
        }

    }  else  { /* unsigned. */

        switch(size)  {
            
          case sizeof(uint32_t):
              string = xmalloc(UINT32_T_STRING_SIZE);
              snprintf(string, UINT32_T_STRING_SIZE, "%"PRIu32"", *(uint32_t *)data);
              break;

          case sizeof(uint16_t):
              string = xmalloc(UINT16_T_STRING_SIZE);
              snprintf(string, UINT16_T_STRING_SIZE, "%"PRIu16"", *(uint16_t *)data);
              break;

          default:
              /* ditto on forced b0rking. */
              fatal_error("illegal size passed to int_to_string()!");
              break;
        }
    }
    
    return string;
}

/* workhorse int to string routine. */
char *network_int_to_string(const unsigned char *data, int data_len,
                            size_t size, int si)
{
    char *string;

    if(size > data_len)
        return NULL;
    
    if(si)  { /* signed. */
        
        switch(size)  {

          case sizeof(int32_t):
              {
                  int32_t tmp;
                  tmp = ntohl(* (int32_t *)data);
                  string = int_to_string((void *)&tmp, data_len, size, si);
                  break;
              }

          case sizeof(int16_t):
              {
                  int16_t tmp;
                  tmp = ntohl(* (int16_t *)data);
                  string = int_to_string((void *)&tmp, data_len, size, si);
                  break;
              }

          default:
             /* we should never get here but we should force ourselves
              * to b0rk if we do. */
             fatal_error("illegal size passed to int_to_string()!");
        }

    }  else  { /* unsigned. */

        switch(size)  {

          case sizeof(uint32_t):
              {
                  uint32_t tmp;
                  tmp = ntohl(* (uint32_t *)data);
                  string = int_to_string((void *)&tmp, data_len, size, si);
                  break;

              }

          case sizeof(uint16_t):
              {
                  uint16_t tmp;
                  tmp = ntohl(* (uint16_t *)data);
                  string = int_to_string((void *)&tmp, data_len, size, si);
                  break;
              }

          default:
              /* ditto on forced b0rking. */
              fatal_error("illegal size passed to int_to_string()!");

        }
    }
    
    return string;

}

/* interface to network_*int*_to_string */
char *network_uint32_to_string(const unsigned char *data, int len)
{
    return(network_int_to_string(data, len, sizeof(uint32_t), 0));
}
char *network_int32_to_string(const unsigned char *data, int len)
{
    return(network_int_to_string(data, len, sizeof(int32_t), 1));
}
char *network_uint16_to_string(const unsigned char *data, int len)
{
    return(network_int_to_string(data, len, sizeof(int16_t), 0));
}
char *network_int16_to_string(const unsigned char *data, int len)
{
    return(network_int_to_string(data, len, sizeof(int16_t), 1));
}

/* interface to *int*_to_string */
char *uint32_to_string(const unsigned char *data, int len)
{
    return(int_to_string(data, len, sizeof(uint32_t), 0));
}
char *int32_to_string(const unsigned char *data, int len)
{
    return(int_to_string(data, len, sizeof(int32_t), 1));
}
char *uint16_to_string(const unsigned char *data, int len)
{
    return(int_to_string(data, len, sizeof(uint16_t), 0));
}
char *int16_to_string(const unsigned char *data, int len)
{
    return(network_int_to_string(data, len, sizeof(int16_t), 1));
}

char *network_uint16_list_to_string(const unsigned char *data, int len)
{
    char *tmp;
    char *s = NULL;
    char *string = NULL;

    if(len == 0 ||
       len%sizeof(uint16_t))
        return NULL;

    for(;len > 0;len -= sizeof(uint16_t))  {

        s = network_uint16_to_string(data, sizeof(uint16_t));
        if(s == NULL)
            return string;

        tmp = string;
        string = splice_string(s, string);
        
        if(tmp != NULL)
            xfree(tmp);
        xfree(s);
        
        tmp = string;
        string = splice_string(string, ";");
        free(tmp);
        
        data += sizeof(uint16_t);
    }
        
    return string;
}

void *string_to_uint16_list(const char *s)
{
    list_t *list_ptr = NULL;
    void *val;
    const char *cp;
    char *ptr;

    cp = s;
    while((ptr = strchr(cp, ';')) != NULL)  {
        *ptr = '\0';
        val = string_to_uint16(cp);
        if(val == NULL)
            break;

        list_ptr = add_to_list(list_ptr,val);
        
        ptr++;
        if(*ptr == '\0')
            break;
        cp = ptr;
    }
    
    return list_ptr;
}

void *string_uint16_list_to_network(const char *s, unsigned char *retlen)
{
    unsigned char *data, *data_ptr;
    uint16_t *val;
    int len = 0;
    char *ptr;
    const char *cp;

    for(cp = s;ptr != NULL;ptr = strchr(cp, ';'))  {
        cp = ptr + 1;
        len++;
    }

    if(len == 0)
        return NULL;

    data = xmalloc(sizeof(uint16_t) * len);
    cp = s;
    data_ptr = data;
    
    while(len--)  {
        ptr = strchr(cp, ';');
        *ptr = '\0';
        
        val = string_to_uint16(cp);
        memcpy(data_ptr, val, sizeof(uint16_t));
        xfree(val);
        data_ptr += sizeof(uint16_t);
    }
    
    return data;
}

/* workhorse string to integer routine. */
/* i'm worried about some of the type issues here. */

static void *string_to_int(const char *s, int si, size_t size)
{
    void *val;

    /* allocate string for str* integer conversion. */
    val = xmalloc(size);
    
    if(si)  { /* signed. */
        
        switch(size)  {

          case sizeof(int32_t):
              *(int32_t *)val = (int32_t)strtol(s, NULL, 10);
              break;

          case sizeof(int16_t):
              *(int16_t *)val = (int16_t)atoi(s); /* this is safe
                                                   * because we're always getting our
                                                   * own serialized values. they should decay fine.
                                                   * don't use this on strings gotten from the network. */
              break;
          default:
            /* we should never get here but we should force ourselves
             * to b0rk if we do. */
            fatal_error("illegal size passed to string_to_int()!");
        }

    }  else  { /* unsigned. */

        switch(size)  {
            
        case sizeof(uint32_t):
            *(uint32_t *)val = (uint32_t)strtoul(s, NULL, 10);
            break;

        case sizeof(uint16_t):
            *(uint16_t *)val = (uint16_t)atoi(s); /* ditto on safeness. */
            break;

        default:
            /* ditto on forced b0rking. */
            fatal_error("illegal size passed to string_to_int()!");
            
        }
    }

    return val;
}

/* Interface for string_to_*int* */
void *string_to_uint32(const char *s)
{
    return(string_to_int(s, 0, sizeof(uint32_t)));
}
void *string_to_int32(const char *s)
{
    return(string_to_int(s, 1, sizeof(uint32_t)));
}
void *string_to_uint16(const char *s)
{
    return(string_to_int(s, 0, sizeof(int16_t)));
}
void *string_to_int16(const char *s)
{
    return(string_to_int(s, 1, sizeof(uint16_t)));
}

/* string int to network. */
static void *string_int_to_network(const char *s, unsigned char *ret_len,
                                  int si, size_t size)
{
    void *val;
    unsigned char *net_val;

    net_val = xmalloc(size);
    *ret_len = size;
    
    if(si)  {
        
        switch(size)  {
            
          case sizeof(int32_t):

            val = string_to_int32(s);
            *(int32_t *)net_val = htonl(*(int32_t *)val);
            xfree(val);
            break;
            
          case sizeof(int16_t):
            val = string_to_int16(s);
            *(int16_t *)net_val = htons(*(int16_t *)val);
            xfree(val);
            break;
            
          default:
            /* we should never get here but we should force ourselves
             * to b0rk if we do. */
            fatal_error("illegal size passed to string_int_to_network()!");
        }

    }  else {
            
        switch(size)  {
        
        case sizeof(uint32_t):

            val = string_to_uint32(s);
            *(uint32_t *)net_val = htonl(*(uint32_t *)val);
            xfree(val);
            break;
            
        case sizeof(uint16_t):

            val = string_to_uint16(s);
            *(uint16_t *)net_val = htons(*(uint16_t *)val);
            xfree(val);
            break;
            
        default:
            /* ditto on the b0rkination. */
            fatal_error("illegal size passed to string_int_to_network()!");
        }
    }

    return net_val;
}

/* interface to string_*int*_to_network */
void *string_int32_to_network(const char *s, unsigned char *ret_len)
{
    return(string_int_to_network(s, ret_len, 1, sizeof(int32_t)));
}
void *string_uint32_to_network(const char *s, unsigned char *ret_len)
{
    return(string_int_to_network(s, ret_len, 0, sizeof(uint32_t)));
}
void *string_int16_to_network(const char *s, unsigned char *ret_len)
{
    return(string_int_to_network(s, ret_len, 1, sizeof(int16_t)));
}
void *string_uint16_to_network(const char *s, unsigned char *ret_len)
{
    return(string_int_to_network(s, ret_len, 0, sizeof(uint16_t)));
}
    
/* variation on network_addr_to_string in dhcp-convert.c we use
 * this when we don't want to run around freeing results up, and
 * know that it will be statically stored. Useful for insertion
 * in message log routines. */

char *network_address_to_string_static(uint32_t addr)
{    
    struct in_addr in;
    in.s_addr =  addr;

    return(inet_ntoa(in));
}

/* Gee, lots of workhorses in this source file. */
