/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-files.c,v 1.12 2002/06/17 13:23:08 actmodern Exp $
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com> 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Handles our variable files.
 *
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-files.h>

static FILE *file_open_proc(const char *filename, char *cmode, int flags, mode_t mode)
{
    FILE *fp;
    int fd;
    
    /* open file and create if it needed. */

    if((fd = open(filename, flags, mode)) < 0)    {
        error_message("cannot open or create file: %s : %s", filename, strerror(errno));
        return NULL;
    }

    if((fp = fdopen(fd, cmode)) == NULL)  {
        close(fd);
        error_message("cannot fdopen: %s : %s", filename, strerror(errno));        
        return NULL;
    }
    
    return fp; /* fd is closed when fp is closed since they share the descriptor. */
}

/* Assumes we're opening a directory only writable by us . */

/* create file or open it if it exists. */
FILE *file_open_or_create_safe(const char *filename, char *mode)
{
    return(file_open_proc(filename, mode, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR));
}

/* create file or open if it exists, and always truncate. */
FILE *file_create_and_truncate_safe(const char *filename, char *mode)
{
    return(file_open_proc(filename, mode, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR));
}

/* delete a file. */
void delete_file(char *fname)
{
    unlink(fname);
}

/* move a file. */
void move_file(char *fname, char *dest)
{
#ifdef HAVE_RENAME
    rename(fname, dest);
#else /* HAVE_RENAME */
    unlink(dest);
    link(fname, dest);
    unlink(fname);
#endif /* HAVE_RENAME */
    return;
}

/* check if a file exists. -- achtung: watch out for race
 * conditions */
int file_exists(char *fname)
{
    FILE *fp;
    
    /* use stdio. fixme: should we stat()?*/
    fp = fopen(fname, "r");
    if(fp != NULL)  {
        fclose(fp);
        return 1;
    } else
        return 0;
}

/* change mode on a file so its publicaly readable. */
void make_file_public_read(char *fname)
{
    chmod(fname, S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
}

/* grab our pid file. <name>.pid */
static char *get_pid_file_name(char *name)
{
    return (splice_string(name, ".pid"));
}

/* create a pid file <name>.pid */
void create_pid_file(char *name)
{
    FILE *fp;
    char *fname;
    
    fname = get_pid_file_name(name);
    fp = file_create_and_truncate_safe(fname, "a");
    if(fp == NULL)
        fatal_error("could not create pid file");
    
    fprintf(fp, "%u", getpid());

    fclose(fp);
    xfree(fname);
    return;
}

/* delete a pid file <name>.pid */
void delete_pid_file(char *name)
{
    char *fname = get_pid_file_name(name);
    
    unlink(fname);
    
    xfree(fname);
    return;
}

/* get pid from file. */
int get_pid_file(char *name, pid_t *pid)
{
    FILE *fp;
    char *fname = get_pid_file_name(name);

    if(!file_exists(fname))  {
        xfree(fname);
        return -1;
    }
    
    fp = file_open_or_create_safe(fname, "r");
    if(fp == NULL)
        fatal_error("could not open pid file");
    
    fscanf(fp, "%u", pid);
    fclose(fp);
    xfree(fname);

    return 0;
}
