/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-globconf.c,v 1.9 2002/06/22 18:11:25 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Global config options -- this breaks our usual order of doing things.
 * Usually we pass the datum instead of keeping a local copy and modifying it.
 * However, in the case of global defaults they are by definition global and should
 * only be accessed from one datum. Thus we keep a local copy we initialize.
 *
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-globconf.h>
#include <dhcp-files.h>
#include <dhcp-convert.h>
#include <dhcp-parser.h>
#include <dhcp-varfile.h>

/* change this before... */
#define GLOBAL_OPTIONS_LEN 3

/* ... changing size of this */
/* array holding information on converting global config options and their defaults.
 * do not change order unless you change defines in dhcp-globconf.h! */
static glob_conf_t global_config_options[] = {
    { "client-dhcp-retries", string_to_int32, int32_to_string, sizeof(int32_t), DHCP_CLIENT_RETRIES_S, NULL },
    { "client-icmp-retries", string_to_int32, int32_to_string, sizeof(int32_t), DHCP_CLIENT_RETRIES_S, NULL },
    { "client-hostname", strdup_wrap, validate_network_string_to_string, 0, NULL, NULL },
};

/* useful for dumping defaults. */
static int glob_conf_dump(const char *filename)
{    
    int i;
    FILE *fp;
    char *var_string;

    fp = file_open_or_create_safe(filename, "w");

    if(fp == NULL)
        return -1;

    for(i = 0;i < GLOBAL_OPTIONS_LEN; i++)  {
        
        if(global_config_options[i].val == NULL)
            continue;

        var_string = global_config_options[i].convert_option_to_string(global_config_options[i].val, global_config_options[i].size);
        if(var_string == NULL)
            fatal_error("could not convert option to string -- this is a bug report it.");

        fprintf(fp,"%s=%s\n", global_config_options[i].name, var_string);
        xfree(var_string);
    }
    
    fclose(fp);
    return 0;
}

static void glob_conf_set_defaults(void)
{
    int i;

    for(i = 0;i < GLOBAL_OPTIONS_LEN;i++) {

        if(global_config_options[i].default_val == NULL)
            continue;

        if(global_config_options[i].val == NULL) {
            global_config_options[i].val = global_config_options[i].convert_option_to_internal(global_config_options[i].default_val);
            if(global_config_options[i].val == NULL)
                fatal_error("glob-conf: unable to convert value for option %s -- default! this is a bug. report it!");
        }
    }

    return;
}

/* read configuration options in and set variables. */
int init_glob_conf(unsigned char *filename)
{
    varfile_t *varfile;
    int i;
    unsigned char match;
    int retval;

    /* we're ok. all config options are in directories only
     * writable by us. */

    if(!file_exists(filename))  { /* dump defaults if not available. */
        glob_conf_set_defaults(); 
        glob_conf_dump(filename);
        return 0;
    }

    varfile = create_varfile(filename, VARFILE_VARVAL_MODE);
    if(varfile == NULL)
        fatal_error("glob-conf: could not open configuration file: %s: %s", filename, strerror(errno));

    while(1) {
        retval = varfile_get_next_var(varfile);

        switch(retval) {


        case VARFILE_SUCCESS:

            match = 0;
            for(i = 0;i < GLOBAL_OPTIONS_LEN;i++) {

                if(string_matches(global_config_options[i].name, varfile_get_name(varfile))) {
                    /* we have a match. */
                    match = 1;

                    /* skip options which have been set, since some of these may be set
                       on the command line. */
                    if(global_config_options[i].val != NULL)
                        break; 

                    global_config_options[i].val = 
                        global_config_options[i].convert_option_to_internal(varfile_get_val(varfile));

                    if(global_config_options[i].val == NULL)
                        fatal_error("glob-conf: unable to convert value for option %s : line no: %d -- munged variable?",
                                    varfile_get_name(varfile), varfile_get_lineno(varfile));
                    break;
                }
            }
          
            if(match != 1)
                warn_message("glob-conf: unable to lookup variable name %s: line no: %d -- skipping", 
                             varfile_get_name(varfile), varfile_get_lineno(varfile));
            break;
    
        case VARFILE_PARSE_ERROR:
            fatal_error("glob-conf: parse error in %s: line %d", 
                        varfile_get_filename(varfile), varfile_get_lineno(varfile));

        case VARFILE_EOF:

            destroy_varfile(varfile);

            /* one more pass to convert defaults if needed.
             * a user could always truncate the file we dumped. */
            glob_conf_set_defaults();
            return 0;

        default:
            fatal_error("glob-conf: internal error -- please report this bug.");
        }
    }

}

void *glob_conf_get_val(int val_index)
{
    return global_config_options[val_index].val;
}

void glob_conf_set_val(void *val, int option)
{
    global_config_options[option].val = global_config_options[option].convert_option_to_internal(val);
}
