/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-net.c,v 1.29 2002/06/23 03:29:13 actmodern Exp $
 *
 * Copyright 2002 Thamer Alharbash
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Raw network module:
 * 
 * The dhcp client or server needs to communicate rawly 
 * through the network device. There are several features
 * we need:
 * 
 * 1) Ability to send and receive ARP packets: (todo)
 * 
 * This is needed for discovery of existing addresses
 * as well as verification that the address assigned
 * to us is unique if it is on the local subnet.
 * 
 * 2) Ability to send and receive ICMP packets: (todo)
 * 
 * We use ICMP echo packets to check if a certain
 * IP address is taken up or not. This is useful
 * only for the server since it may need to verify 
 * that hosts on a different subnet are alive.
 * 
 * 3) Ability to send and receive DHCP packets:
 * 
 * We need to be able to send and receive
 * DHCP packets in order to communicate
 * with DHCP servers or clients.
 * 
 * 4) Modification of ethernet headers:
 * 
 * When we send broadcast packets we need
 * to set the destination ethernet header
 * accordingly.
 *
 * The only limitation here is that we can only 
 * do DHCP over ethernet which is fine -- for now.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-rtt.h>
#include <dhcp-net.h>

#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/bpf.h>

/* get interface address. */
int rawnet_get_hw_addr(rawnet_t *net, eth_addr_t *addr)
{
    /* are we using a fake address? */
    if(net->fake_hw_addr != NULL) {
        memcpy(addr, net->fake_hw_addr->data, ETH_ADDR_LEN);
        return 0;
    } else {
        return(eth_get(net->eth, addr));
    }
}

/* get real interface address skipping fake even if its available. */
int rawnet_get_real_hw_addr(rawnet_t *net, eth_addr_t *addr)
{
    return(eth_get(net->eth, addr));
}

/* get IP address. */
int rawnet_get_ip_addr(rawnet_t *net, ip_addr_t *addr)
{
    return(interface_get_ip_addr(net->intf_handle, addr));
}

/*
 * Initialization of packet capturing device.
 * 
 */

static pcap_t *initialize_pcap_device(char *interface_name,
                                      char *filter_string,
                                      int promiscuous)
{
    pcap_t *pcap;
    struct bpf_program filter;    
    char errbuf[PCAP_ERRBUF_SIZE];
    uint32_t mask = 0xffffffff;

#if defined(HAVE_BPF_IMMEDIATE)
    int fd;
    u_int on = 1;
#endif /* HAVE_BPF_IMMEDIATE */

    /* If interface not specificed get
     * a suitable one from pcap_lookupdev */

    if(interface_name == NULL)
	interface_name = pcap_lookupdev(errbuf);

    if(interface_name == NULL)  {
        error_message("rawnet: pcap_lookupdev: %s", errbuf);
        return NULL;
    }

    /* Open packet capturing device. */

    pcap = pcap_open_live(interface_name, SNAPLEN, promiscuous, 0, errbuf);
    if(pcap == NULL)  {
        error_message("rawnet: pcap_openlive: %s", errbuf);
        return NULL;
    }

    if(pcap_compile(pcap, &filter, filter_string, 1, mask) == -1)  {
        pcap_close(pcap);
        error_message("rawnet: pcap_compile: %s",pcap_geterr(pcap));
        return NULL;
    }

    if(pcap_setfilter(pcap, &filter) == -1)  {
        pcap_close(pcap);
        error_message("rawnet: pcap_setfilter: %s", pcap_geterr(pcap));
        return NULL;
    } 

    pcap_freecode(&filter);

#if defined(HAVE_BPF_IMMEDIATE)
    /* FreeBSD (others too?) need us to set the descriptor non
     * blocking and do a bogus read. This starts the timer. */

    fd = pcap_fileno(pcap);
    ioctl(fd, BIOCIMMEDIATE, &on);

#endif /* HAVE_BPF_IMMEDIATE */
    return pcap;
}

/* wrapper around interface handler routines. */

/* bring up interface, and assign ourselves the address passed. */
int rawnet_interface_up(rawnet_t *net, ip_addr_t addr, ip_addr_t netmask, int mtu)
{

    if(interface_up(net->intf_handle, addr, netmask, mtu)) {
        error_message("rawnet: could not bring up interface: %s: %s", net->device, strerror(errno));
        rawnet_destroy(net);
        return 1;
    }

    memcpy(&net->ip_addr, &addr, IP_ADDR_LEN); /* set our IP address. */
    return 0;
}

/* bring down interface. */
int rawnet_interface_down(rawnet_t *net)
{
    if(interface_down(net->intf_handle)) {
        error_message("rawnet: could not bring down interface: %s: %s", net->device, strerror(errno));
        rawnet_destroy(net);
        return 1;
    }

    memset(&net->ip_addr, 0, IP_ADDR_LEN); /* clear our IP address -- this isn't very useful though. */
    return 0;
}

/* Create the workhorse network module. */

rawnet_t *rawnet_create(char *device, char *filter, 
                        int dhcp_src_port, int dhcp_dst_port,
                        int promiscuous)
{
    rawnet_t *net;

    net = xcalloc(sizeof(rawnet_t));
    net->src_port = dhcp_src_port;
    net->dst_port = dhcp_dst_port;
    net->device = strdup(device);
    net->pcap_filter = strdup(filter);
    net->packet_data = xcalloc(DEFAULT_MTU);  /* XXX -- fixme make mtu configurable. */
    net->promiscuous = promiscuous;

    /* get the interface handle for bringing it up or down and querying it. */
    net->intf_handle = create_interface_control(net->device);
    if(net->intf_handle == NULL) {
        rawnet_destroy(net);
        return NULL;
    }

    /* Is the interface up? */
    if(!interface_is_up(net->intf_handle)) {
        /* then bring it up with dum values. */
        if(rawnet_interface_up(net, 0, 0xffffff, -1))  { 
            rawnet_destroy(net);
            return NULL;
        }
    }

    /* Get the interface and make sure its an ethernet interface or
     * else we're hosed. */

    net->eth = eth_open(device);
    if(net->eth == NULL)  {
        error_message("rawnet: device not supported or unavailable: %s: %s", device, strerror(errno));
        rawnet_destroy(net);
        return NULL;
    }

    /* get pcap handler */
    net->pcap = initialize_pcap_device(device, filter, promiscuous);
    if(net->pcap == NULL)  {
        rawnet_destroy(net);
        return NULL;
    }

    net->pcap_fd = pcap_fileno(net->pcap);
    if(net->pcap_fd < 0)  {
        error_message("rawnet: cannot get descriptor for %s, this system may is not supported.", device);
        rawnet_destroy(net);
        return NULL;
    }

    /* When we first initialize we use a real hardware
     * address. This can be changed later to a fake hardware
     * address if the system supports it. */

    rawnet_get_hw_addr(net, &net->hw_addr);
    rawnet_get_ip_addr(net, &net->ip_addr);

    /* Packet handler objects.
     * 
     * These exit on failure.
     * 
     */

    net->ether_p = eth_create();
    net->ip_p = ip_create();
    net->arp_p = arp_create();
    net->icmp_p = icmp_create();
    net->udp_p = udp_create();
    net->dhcp_p = dhcp_create();

    return net;
}

/* Use a fake hardware address (diagnostic purposes).  We need to
 * reinitialize our rawnet to use promiscuous packet capturing
 * since we do want to accept packets for this hardware
 * address. */

void rawnet_use_fake_hw_addr(rawnet_t *net, char *mac_string)
{
    unsigned char mac_addr[ETH_ADDR_LEN];

    net->fake_hw_addr = xmalloc(sizeof(eth_addr_t));
    if(hex_string_to_value(mac_string, mac_addr))
        fatal_error("malformed mac_string for fake client hw addr. please check mac address.");
    memcpy(net->hw_addr.data, mac_addr, ETH_ADDR_LEN);

    /* now reinitialize in promiscuous mode */
    reinitialize_rawnet(net, 1);

    return;
}
        
void rawnet_destroy(rawnet_t *net)
{
    if(net->eth != NULL)
        eth_close(net->eth);

    if(net->intf_handle)
        destroy_interface_control(net->intf_handle);
    
    if(net->pcap != NULL)
        pcap_close(net->pcap);
    
    if(net->device)
        xfree(net->device);

    if(net->packet_data)
        xfree(net->packet_data);
    
    if(net->pcap_filter)
        xfree(net->pcap_filter);

    if(net->ether_p)
        eth_destroy(net->ether_p);
    if(net->ip_p)
        ip_destroy(net->ip_p);
    if(net->arp_p)
        arp_destroy(net->arp_p);
    if(net->icmp_p)
        icmp_destroy(net->icmp_p);
    if(net->udp_p)
        udp_destroy(net->udp_p);
    if(net->dhcp_p)
        dhcp_destroy(net->dhcp_p);

    xfree(net);
    
    return;
}

/* The rawnet up and down routines don't mess with the interface
 * intentionally. They're meant only to close our packet capturing
 * and sending devices so as to free up system resources. */

/* Bring pcap/dnet up and down.
 * Useful for wait states when
 * we don't want our packet buffers
 * to fill up. */

void rawnet_down(rawnet_t *net)
{
    pcap_close(net->pcap);
    net->pcap = NULL;
    eth_close(net->eth);
    net->eth = NULL;

    return;
}

int rawnet_up(rawnet_t *net)
{
    net->pcap = initialize_pcap_device(net->device, net->pcap_filter, net->promiscuous);
    if(net->pcap == NULL)  {
        return -1;
    }
    net->pcap_fd = pcap_fileno(net->pcap);
    
    net->eth = eth_open(net->device);
    if(net->eth == NULL)
        return -1;
    
    return 0;
}

void reinitialize_rawnet(rawnet_t *net, int promiscuous)
{
    net->promiscuous = promiscuous;
    rawnet_down(net);
    rawnet_up(net);

    return;
}

/* 
 * rawnet get packet:
 * we already expect a filter to only allow
 * the following types of packets:
 * ARP, UDP, and ICMP.
 * 
 * Anything else is considered malformed.
 * 
 */

int rawnet_get_packet(rawnet_t *net, struct timeval *tm)
{
    const unsigned char *packet;
    struct pcap_pkthdr pkthdr;
    int len;
    fd_set read_set;
    int retval;

    net->type = RAWNET_UNKNOWN;

    while(1) {
        FD_ZERO(&read_set);
        FD_SET(net->pcap_fd, &read_set);
    
        retval = select((net->pcap_fd + 1), &read_set, NULL, NULL, tm);
        
        if(retval == 0)
            return RAWNET_TIMEOUT;
        
        if(retval < 0)
            return RAWNET_PCAP_ERROR;

        packet = pcap_next(net->pcap, &pkthdr);
        if(packet == NULL)
            continue; /* BPF may return 1 when there really isn't any packet for us.
                       * is this because we're notified prior to filtering?
                       * pcap only returns NULL on EWOULDBLOCK for bpf -- even though
                       * in debugging I *never* get that errno value :| */
        else
            break;
    }

    /* Set timestamp. */

#if defined(HAVE_BPF_TIMEVAL)

    /* This is an ugly hack -- why oh why did OpenBSD insist on
     * creating its own bpf_timeval struct with unsigned values
     * as opposed to the common UNIX timeval which has signed values.
     *
     *
     * *sigh*
     *
     * There really isn't any clean way around this apart from grabbing
     * our own timestamp instead of using the bpf one. This is inaccurate
     * but until OpenBSD fixes their braindamage they'll just have to
     * deal with the inaccuracy. Anything else would be too damaging
     * to the other UNIX flavors.
     */

    gettimeofday(&net->tstamp, NULL);

#else /* HAVE_BPF_TIMEVAL */

    net->tstamp = pkthdr.ts;

#endif /* HAVE_BPF_TIMEVAL */

    /* Read ethernet header and switch on type.   *
     * Fails if malformed or insufficient length. */

    if(eth_read_packet_image(net->ether_p, packet, pkthdr.caplen) < 0)
        return RAWNET_MALFORMED_PACKET;
    
    packet += ETH_HDR_LEN;
    pkthdr.caplen -= ETH_HDR_LEN;

    switch(eth_get_type(net->ether_p))  {
        
      case ETH_TYPE_ARP:
          if(arp_read_packet_image(net->arp_p, packet, pkthdr.caplen) < 0)
              return RAWNET_MALFORMED_PACKET;

          net->type = RAWNET_ARP;
          return RAWNET_OK;
        
      case ETH_TYPE_IP:
        if(ip_read_packet_image(net->ip_p, packet, pkthdr.caplen) < 0)
            return RAWNET_MALFORMED_PACKET;

        len = ip_get_hl(net->ip_p);
        packet += len;
        pkthdr.caplen -= len;

        switch(ip_get_proto(net->ip_p))  {
            
          case IP_PROTO_UDP:
            if(udp_read_packet_image(net->udp_p, packet, pkthdr.caplen) < 0)
                return RAWNET_MALFORMED_PACKET;
            
            /* We assume any UDP packet does contain a DHCP packet.
             * So try and grab it here. */
            
            packet += UDP_HDR_LEN;
            pkthdr.caplen -= UDP_HDR_LEN;
            
            /* purge dhcp so we get rid of old options. */
            dhcp_purge(net->dhcp_p);

            if(dhcp_read_packet_image(net->dhcp_p, packet, pkthdr.caplen) < 0)
                return RAWNET_MALFORMED_PACKET;
            
            net->type = RAWNET_DHCP;
            return RAWNET_OK;
            
          case IP_PROTO_ICMP:
            if(icmp_read_packet_image(net->icmp_p, packet, pkthdr.caplen) < 0)
                return RAWNET_MALFORMED_PACKET;

            net->type = RAWNET_ICMP;
            return RAWNET_OK;

        default:
            break;
        }
        
      default:
        break;

    }

    return RAWNET_UNHANDLED;
}

int rawnet_send_packet(rawnet_t *net)
{
    if(eth_send(net->eth, net->packet_data, net->packet_len) != net->packet_len)
        return 1;
    else
        return 0;
}

char *rawnet_strerror(rawnet_t *net)
{
    return(pcap_geterr(net->pcap));
}


/* Just update the seconds field.
 * 
 * FIXME: mod this so that it hooks
 * into the rest of the framework.
 * right now it's error prone.
 */

void rawnet_dhcp_update(rawnet_t *net,
                        time_t seconds)
{
    unsigned char *packet;
    uint16_t secs = htons(seconds);
    
    /* FIX ME: we're assuming the ip packet header
     * has no options. 
     */

    packet = (net->packet_data + ETH_HDR_LEN + IP_HDR_LEN + 
              UDP_HDR_LEN);

    packet += 8; /* to seconds field. */
    memcpy(packet, &secs, 2);
    
    ip_checksum((net->packet_data + ETH_HDR_LEN), (net->packet_len - ETH_HDR_LEN));

    return;
}

/* 
 * We can ask rawnet to validate for us.
 * Insert checks here to make sure
 * the packet is as kosher as christmas
 * (chicken soup).
 * 
 */

int rawnet_is_valid(rawnet_t *net)
{
    switch(net->type)  {
        
       case RAWNET_DHCP:
         if(!dhcp_valid_magic_cookie(net->dhcp_p))
           break;
        else
          return 1;
        
       default: /* todo: we need more validation on other packets. */
        break;
    }

    return 0;
}

/* Generic send and wait for reply:
 *
 * Calls update() after timeout.
 * Calls check() on every packet coming in.
 *
 */

int rawnet_packet_transact(rawnet_t *net, void *arg, void (*update)(void *arg),
                           int (*check)(void *arg))
{
    rtt_t *rtt;
    int retval;
    unsigned char send_packet = 1;
    struct timeval timeout;

    /* create rtt mechanism. */
    rtt = rtt_create();

    /* We loop here and re-send packets if we timeout, or fail a
     * check on an incoming packets. */

    while(1)  {

        if(send_packet)  {
            
            send_packet = 0;

            if(rawnet_send_packet(net) < 0)  {
                error_message("rawnet: could not send packet: %s", strerror(errno));
                rtt_destroy(rtt);                
                return -1;
            }

        }

        timeout = rtt_get_timeout(rtt);
        retval = rawnet_get_packet(net, &timeout);

        switch(retval)  {

          case RAWNET_PCAP_ERROR:
            error_message("rawnet: couldn't wait for packets: %s : %s", net->device,
                          strerror(errno));
            rtt_destroy(rtt);
            return -1;

          case RAWNET_TIMEOUT:

            rtt_timeout(rtt);
            if(!rtt_can_retry(rtt))  {
                rtt_destroy(rtt);
                return -1;
            }

            if(update) /* update if needed. only called after timeout. */
                update(arg);

            send_packet = 1;
            break;
            
          case RAWNET_OK:

              if(!check) /* no check? return 0 anyway (we want
                          * any packet it seems). */
                  return 0;
              
              if(check(arg))  {  /* if we have a check 
                                  * then check packet. */
                  rtt_destroy(rtt);
                  return 0; /* it's good. let's return. */
              }
              
              /* otherwise we should just continue. */
              break;

        case RAWNET_UNHANDLED:
        case RAWNET_MALFORMED_PACKET:
            break;

        default:
            error_message("rawnet: invalid return from rawnet_get_packet() -- this is a bug report it please");
            break;
        }
    }
}
