/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-parser.c,v 1.2 2002/06/17 13:23:08 actmodern Exp $
 *
 * Copyright 2002 Thamer Alharbash
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Parser driver. Here we can build custom parsers by plugging in
 * the applicable functions for each single character token.  A
 * default parser with a 0 for a character is used as a fallback
 * (and to mark the end of the parser driver)
 *
 * This parser will b0rk on something that's not printable and not even
 * call the parse function.
 *
 * We build from the parser utility routines or we can pass our
 * own.
 *
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>
#include <dhcp-stringbuffer.h>
#include <dhcp-files.h>
#include <dhcp-parser.h>

static int is_valid_string_char(int c)
{
    if(ispunct(c) || isalnum(c))
        return 1;
     else
         return 0;
}

static token_t check_eof(parser_t *input)
{
    if(feof(input->fp))
        return PARSER_TOKEN_EOF;
    else
        return PARSER_TOKEN_ERROR;
}

/* just gobble up till the end of the newline */
static token_t parse_gobble_line(parser_t *input)
{
    int c;

    while(1) {
        c = fgetc(input->fp);

        switch(c) {

        case EOF:
            return check_eof(input);

        case '\n':
            input->line_no++;
            return PARSER_TOKEN_NEWLINE;

        default:
            break;
        }
    }
}

static token_t parse_get_quoted_string(parser_t *input)
{
    int c;

    while(1) {

        c = fgetc(input->fp);

        switch(c) {

        case EOF:
            return check_eof(input);

        case '\"':
            return PARSER_TOKEN_QUOTED_STRING;

        case '\n':
            return PARSER_TOKEN_ERROR; /* error to have newline in string constant. */

        case ' ':
        case '\t':
        default:
            if(is_valid_string_char(c) || c == ' ' || c == '\t')
                stringbuffer_append_c(input->data_buff, (char)c);
            else {
                ungetc(c, input->fp);
                return PARSER_TOKEN_ERROR;
            }
        }
    }
}

static token_t parse_get_string(parser_t *input)
{
     int c;

    while(1) {

        c = fgetc(input->fp);

        switch(c) {

        case EOF:
            return check_eof(input);

        case '\n':
            ungetc(c, input->fp);
            return PARSER_TOKEN_STRING; /* newline terminates string. */

        case '\t':
        case ' ':
            ungetc(c, input->fp);
            return PARSER_TOKEN_STRING; /* newline terminates string. */

        case '=':
            ungetc(c, input->fp);
            return PARSER_TOKEN_STRING; /* newline terminates string. */

        default:
            if(is_valid_string_char(c))
                stringbuffer_append_c(input->data_buff, (char)c);
            else {
                ungetc(c, input->fp);
                return PARSER_TOKEN_ERROR; /* anything else and we've been terminated. */
            }
        }
    }
}

atom_t get_next_atom(parser_t *input)
{
    int c;
    token_t token;

    stringbuffer_clear(input->data_buff);

    while(1) {
        
        c = fgetc(input->fp);

        switch(c) {

        case EOF:
            return PARSER_ATOM_EOF;

        case '#':

            token = parse_gobble_line(input);

            if(token == PARSER_TOKEN_NEWLINE ||
               token == PARSER_TOKEN_EOF)
                return PARSER_ATOM_END_OF_STATEMENT;

            /* anything else is an error. */
            return PARSER_ATOM_ERROR;

        case '\n':
            input->line_no++;
            return PARSER_ATOM_END_OF_STATEMENT;

        case '=':
            return PARSER_ATOM_ASSIGNMENT;

        case '\t':
        case ' ':
            break;

        case '\"':
            token = parse_get_quoted_string(input);

            if(token == PARSER_TOKEN_ERROR)
                return PARSER_ATOM_ERROR;

            return PARSER_ATOM_STRING;

        default:
            /* if we have an alphanumeric it's a string.
               read it in. */
            if(is_valid_string_char(c)) {

                ungetc(c, input->fp);
                token = parse_get_string(input);

                if(token == PARSER_TOKEN_ERROR)
                    return PARSER_ATOM_ERROR;

                return PARSER_ATOM_STRING;
            }
            return PARSER_ATOM_ERROR;
        }
    }
}

parser_t *create_parser(const char *filename)
{
    FILE *fp;
    parser_t *parser;

    fp = file_open_or_create_safe(filename, "r");
    if(fp == NULL)
        return NULL;

    parser = xmalloc(sizeof(parser_t));
    parser->fp = fp;
    parser->line_no = 1;
    parser->data_buff = create_stringbuffer();

    return parser;
}

void destroy_parser(parser_t *parser)
{
    fclose(parser->fp);
    destroy_stringbuffer(parser->data_buff);
    xfree(parser);

    return;
}

int parser_get_line_no(parser_t *parser)
{
    return parser->line_no;
}

const char *parser_get_data(parser_t *parser)
{
    return stringbuffer_getstring(parser->data_buff);
}
