/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-stringbuffer.c,v 1.6 2002/06/22 18:11:25 actmodern Exp $
 *
 * Copyright 2002 Thamer Alharbash
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>

/* * * * * * * * * * * * *
 * raw string routines.  *
 * * * * * * * * * * * * */

/* just malloc out a string. */
static char *allocate_string(int len)
{
    char *s;
    s = xmalloc(sizeof(char) * len + 1); /* add for null termination. */
    s[len] = 0;
    return s;
}

/* extend a string. */
static char *extend_string(char *str, int cur_len, int ex_len)
{
    int len = strlen(str);

    str = realloc(str, (sizeof(char) * (len + ex_len + 1)));
    str[len + ex_len] = 0; /* make sure it's null terminated. */

    return str;
}

/* get a substring. */
static char *substring(char *begin, int len)
{
    char *new_string;

    new_string = allocate_string(len);
    strncpy(new_string, begin, len);
    new_string[len] = 0;

    return new_string;
}

/* used in aligning -- we try to get words up to end. */
static char *get_string_align(char *s, int end, int *len)
{
    char *cur_ptr;

    if(s == 0 || *s == 0) /* end of string or no string. */
        return NULL;

    /* if strlen is smaller than len go ahead and just return it. */
    if(strlen(s) < end) {
        *len = strlen(s);
        return strdup(s);
    }

    /* otherwise we need to hop to len */
    cur_ptr = &s[end - 1];

    /* now check to see if we have a whitespace behind cur_ptr
     * if we do return at that point. */

    for(;cur_ptr != s;cur_ptr--) {
        if(*cur_ptr == ' ' ||
           *cur_ptr == '\t') {
            /* copy here and return. */
            *len = (cur_ptr - s) + 1;
            return(substring(s, *len));
        }
    }

    /* keep walking till we find whitspace or end. */
    for(cur_ptr = &s[end - 1];*cur_ptr != 0 && *cur_ptr != ' ' && *cur_ptr != '\t';cur_ptr++);

    *len = (cur_ptr - s) + 1;
    return(substring(s, *len));
}

/* * * * * * * * * * * * * * *
 * stringbuffer routines.    *
 * * * * * * * * * * * * * * */

/* create a new stringbuffer */
stringbuffer *create_stringbuffer(void)
{
  stringbuffer *sb;

  sb = xmalloc(sizeof(stringbuffer));
  sb->len = 0;
  sb->capacity = 0;
  sb->buf = allocate_string(0);

  return sb;
}

/* destroy the stringbuffer */
void destroy_stringbuffer(stringbuffer *sb)
{
  free(sb->buf);
  free(sb);
}

/* clear a string. */
void stringbuffer_clear(stringbuffer *sb)
{
    sb->len = 0;
    sb->buf[sb->len] = 0;
}

/* append character to stringbuffer */
void stringbuffer_append_c(stringbuffer *sb, char c)
{
  if(sb->capacity <= (1 + sb->len)) {
      sb->buf = extend_string(sb->buf, sb->len, 1);
      sb->len++;
      sb->capacity = sb->len;
  } else
      sb->len++;

  sb->buf[sb->len - 1] = c;
  sb->buf[sb->len] = 0;
}

/* append string to stringbuffer */
void stringbuffer_append(stringbuffer *sb, const char *s)
{
  int len = strlen(s);

  if(sb->capacity <= (len + sb->len)) {
      sb->buf = extend_string(sb->buf, sb->len, len);
      sb->len += len;
      sb->capacity = sb->len;
  } else
      sb->len += len;

  strcat(sb->buf, s);
  sb->buf[sb->len] = 0;
}

/* remove whitespace */
stringbuffer *stringbuffer_trim_whitespace(stringbuffer *sb)
{
  char *ptr, *curptr;
  char *end, *begin;
  char *newbuf;
  int len;

  if(sb->len == 0) /* empty string. */
    return sb;

  for(begin = &sb->buf[0];
      *begin != '\0' && (*begin == ' ' || *begin == '\t');
      begin++);

  if(*begin != '\0')  { 

    /* we do have whitespace in the beginning so find the end. */
    for(end = &sb->buf[sb->len];
	*end == ' ' || *end == '\t';
	end--);

    len = sb->len; /* save old length. */
    /* create a new string */
    newbuf = allocate_string(len);

      ptr = newbuf;
      for(curptr = begin;curptr != end;curptr++) {
	*ptr = *curptr;
	ptr++;
      }
      *ptr = 0;

      xfree(sb->buf);
      sb->buf = newbuf;
      sb->len = strlen(newbuf);
      sb->capacity = len;

   
  } else {

      /* zap beginning of string. since its all whitespace. */
      sb->buf[0] = 0;
      sb->len = 0;
  }

  return sb;
}

/* get the last occurance of a specific character. useful for slicing. */
char *stringbuffer_get_last_occurance(stringbuffer *sb, char c)
{
  char *ptr, *ptrend = NULL;

  ptr = sb->buf;
  while((ptr = strchr(ptr, '\n')) != NULL) {
    ptrend = ptr;
    ptr++;
  }

  return ptrend;
}

/* remove the last newline character. */
void stringbuffer_trim_newline(stringbuffer *sb)
{
  char *ptr;

  ptr = stringbuffer_get_last_occurance(sb, '\n');
  if(ptr != NULL)
    *ptr = 0;


  ptr = stringbuffer_get_last_occurance(sb, '\r');
  if(ptr != NULL)
    *ptr = 0;


  sb->len = strlen(sb->buf);

  return;
}

/* return the C string from the buffer. */
const char *stringbuffer_getstring(stringbuffer *sb)
{
  return sb->buf;
}

/* set a string into a stringbuffer */
void stringbuffer_set(stringbuffer *dest, const char *s)
{
    stringbuffer_clear(dest); /* zap */
    stringbuffer_append(dest, s);
}

/* copy a stringbuffer into another stringbuffer */
void stringbuffer_copy(stringbuffer *dest, stringbuffer *src)
{
    stringbuffer_set(dest, stringbuffer_getstring(src));
}

/* replace in stringbuffer occurances of c with replace */
void stringbuffer_replace_c(stringbuffer *sb, char c, char replace)
{
    char *ptr;

    for(ptr = sb->buf; *ptr != 0; ptr++) {
        if(*ptr == c)
            *ptr = replace;
    }
}

/* zap newlines by placing spaces in their place.
 * we use this before aligning. */
static void stringbuffer_zap_newline(stringbuffer *sb)
{

    stringbuffer_replace_c(sb, '\n', ' ');
    stringbuffer_replace_c(sb, '\r', ' ');
}

/* align a stringbuffer on begin and end columns. */
void stringbuffer_align(stringbuffer *sb, int begin, int end)
{
    char *ptr, *word_string;
    stringbuffer *aligned_string;
    int len, i;


    stringbuffer_zap_newline(sb);

    aligned_string = create_stringbuffer();
    ptr = sb->buf;

    while(1) {

        word_string = get_string_align(ptr, end, &len);

        if(word_string == NULL)
            break;

        ptr += len;

        for(i = 0;i < begin;i++)
            stringbuffer_append(aligned_string, " ");

        stringbuffer_append(aligned_string, word_string);
        stringbuffer_append(aligned_string, "\n");
        xfree(word_string);

    }

    stringbuffer_copy(sb, aligned_string);
    destroy_stringbuffer(aligned_string);

    return;
}

/* stringbuffer_*printf* these all use snprintf/snvprintf internally */

/* append vprintf with alignment. */
void stringbuffer_avprintf_align(stringbuffer *sb, int start, int end, char *fmt, va_list ap)
{
    stringbuffer *tmp_sb;
    char *str;
    int total, len;

    /* our first malloc is bogus. */
    len = 1;
    str = xmalloc(sizeof(char) * len);
    total = vsnprintf(str, len, fmt, ap);

    /* total is the real length needed. */
    xfree(str);
    len = total + 1;
    str = xmalloc(sizeof(char) * len);
    vsnprintf(str, len, fmt, ap);

    /* now align if we want to align. */
    if(start != 0 && end != 0) {

        tmp_sb = create_stringbuffer();

        stringbuffer_append(tmp_sb, str);
        stringbuffer_align(tmp_sb, start, end);
        stringbuffer_append(sb, stringbuffer_getstring(tmp_sb));

        destroy_stringbuffer(tmp_sb);

    } else {
        stringbuffer_append(sb, str);
    }

    xfree(str);

    return;
}

/* append printf. */
void stringbuffer_aprintf(stringbuffer *sb, char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);

    stringbuffer_avprintf(sb, fmt, ap);

    va_end(ap);
}

/* append printf with alignment. */
void stringbuffer_aprintf_align(stringbuffer *sb, int start, int end, char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);

    stringbuffer_avprintf_align(sb, start, end, fmt, ap);

    va_end(ap);
}

/* append vprintf. */
void stringbuffer_avprintf(stringbuffer *sb, char *fmt, va_list ap)
{
    stringbuffer_avprintf_align(sb, 0, 0, fmt, ap);
}

/* mark newlines to walk through. 
 * our sentinel is the null terminator.
 * two null terminations marks the end of the string.
 * we're guaranteed it is a unique sentinel since
 * we never accept null terminators from outside sources
 * and never build our own strings (obviously!) with
 * null terminators inside of them.
 */
void stringbuffer_marknewlines(stringbuffer *sb)
{
    char *c;

    /* first append one null termination to the end
     * to act as a proper terminator. */
    stringbuffer_append_c(sb, 0);


    c = sb->buf;
    while(1) {

        if(*c == '\n')
            *c = 0;

        c++;
        if(*c == 0)
            break;
    }

    return;
}

/* called _after_ newlines are marked. */
char *stringbuffer_getnextline(stringbuffer *sb, char *c)
{
    char *ptr;

    if(c == NULL) {

        /* get first line. */
        c = sb->buf;

    } else {

        /* check for end. */
        if(*c == 0 && (*(c + 1)) == 0)
            return NULL;

        else {
            /* otherwise move to next line and return */
            c++;
            for(ptr = c;*ptr != 0;ptr++);
            c = ptr + 1;
        }
    }

    return c;
}
