/* $Header: /cvsroot/dhcp-agent/dhcp-agent/dhcp-util.c,v 1.20 2002/06/22 18:11:25 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Utility routines.
 * 
 */

#include <dhcp-agent.h>
#include <dhcp-util.h>

static char msgbuff[MSG_BUFFER_SIZE];
static int verbosity_level = WARNING_VERBOSITY_LEVEL;

/* set the verbosity level. */
int set_verbosity_level(int verbosity_level_to_set)
{
    if(verbosity_level_to_set < 0 || verbosity_level_to_set > MAX_VERBOSITY_LEVEL)
        return 1;
    
    verbosity_level = verbosity_level_to_set;
    return 0;
}

/* get the current verbosity level. */
int get_verbosity_level(void)
{
    return verbosity_level;
}

/* send error message. */
void error_message(char *fmt, ...)
{
    va_list ap;

    if(verbosity_level == QUIET_VERBOSITY_LEVEL)
        return;
    
    va_start(ap, fmt);

    vsnprintf(msgbuff, sizeof(msgbuff), fmt, ap);

    if(interactive == 1)  {
        fprintf(stderr, "%s: error: %s", getprogname(), msgbuff);
        fprintf(stderr,"\n");
    } else
        error_log(msgbuff);

    va_end(ap);
}

/* send error message and exit. (convenience) */
void fatal_error(char *fmt, ...)
{
    va_list ap;

    if(verbosity_level == QUIET_VERBOSITY_LEVEL)
        return;

    va_start(ap, fmt);

    vsnprintf(msgbuff, sizeof(msgbuff), fmt, ap);
    
    if(interactive == 1)  {
        fprintf(stderr, "%s: fatal error: %s", getprogname(), msgbuff);
        fprintf(stderr,"\n");
    } else
        error_log(msgbuff);

    exit(1);

}

/* send error message and exit. (convenience) */
void fatal_error_exception(stringbuffer *exception_string)
{
    char *c = NULL;
    if(verbosity_level == QUIET_VERBOSITY_LEVEL)
        return;
    
    stringbuffer_marknewlines(exception_string);

    if(interactive == 1)
        fprintf(stderr, "%s: fatal error\n", getprogname());
    
    while(1) {

        c = stringbuffer_getnextline(exception_string, c);
        if(c == NULL)
            break;

    if(interactive == 1)  {
        fprintf(stderr, "%s: %s", getprogname(), c);
        fprintf(stderr,"\n");
    } else
        error_log(c);

    }

    exit(1);
}

/* send info message */
void info_message(char *fmt, ...)
{
    va_list ap;

    if(!(verbosity_level >= NORMAL_VERBOSITY_LEVEL))
        return;
    
    va_start(ap, fmt);

    vsnprintf(msgbuff, sizeof(msgbuff), fmt, ap);
    
    if(interactive == 1)  {
        fprintf(stdout, "%s: %s", getprogname(), msgbuff);
        fprintf(stdout,"\n");
    } else
        info_log(msgbuff);

    va_end(ap);
}

/* send warning messages: includes diagnostics. */
void warn_message(char *fmt, ...)
{
    va_list ap;

    if(!(verbosity_level >= WARNING_VERBOSITY_LEVEL))
        return;
    
    va_start(ap, fmt);

    vsnprintf(msgbuff, sizeof(msgbuff), fmt, ap);
    
    if(interactive == 1)  {
        fprintf(stdout, "%s warning: %s", getprogname(), msgbuff);
        fprintf(stdout,"\n");
    } else
        info_log(msgbuff);

    va_end(ap);
}

/* send debug message. */
void debug_message(char *fmt, ...)
{
    va_list ap;

    if(!(verbosity_level >= DEBUG_VERBOSITY_LEVEL))
        return;
    
    va_start(ap, fmt);

    vsnprintf(msgbuff, sizeof(msgbuff), fmt, ap);
    
    if(interactive == 1)  {
        fprintf(stdout, "%s: debug: %s", getprogname(), msgbuff);
        fprintf(stdout,"\n");
    } else
        info_log(msgbuff);

    va_end(ap);
}

/* 
 * Malloc and free wrapper functions in case
 * we want to do some accounting or debugging here.
 * It also calls fatal_error aborting us if
 * we ever run out of memory.
 * 
 */

void *xmalloc(size_t size)
{
    void *p;

    p = malloc(size);
    if(p == NULL)
	fatal_error("could not allocate memory! aborting!");

    return p;
}

void *xcalloc(size_t size)
{
    void *p;
    
#ifdef HAVE_CALLOC
    p = calloc(1, size);
    if(p == NULL)
        fatal_error("could not allocate memory! aborting!");
#else
    p = xmalloc(size); /* error handled by xmalloc */
    memset(p, 0, size);
#endif
    
    return p;
}

void xfree(void *p)
{
    free(p);
    return;
}

#ifndef HAVE_STRDUP

char *strdup(char *s)
{
    char *newstring;
    int len = strlen(s);

    newstring = xmalloc(len + 1);
    strncpy(newstring, s, len);
    newstring[len] = '\0';

    return newstring;
}

#endif

/* signal manipulation routines. */

/* add signal handler. */
void add_interrupt_handler(int sigtype, void (*sighandler)(int))
{
    struct sigaction sa;
    
    memset(&sa, 0, sizeof(struct sigaction));
    
    sa.sa_handler = sighandler;
    sigfillset(&sa.sa_mask);
    
    if(sigaction(sigtype, &sa, NULL))
        fatal_error("could not set signal handler for: %d", sigtype);
        
    return;
}

/* block all signals. */
void block_interrupts(void)
{
    sigset_t signal_mask;
    
    sigfillset(&signal_mask);
    
    sigprocmask(SIG_SETMASK, &signal_mask, NULL);
    
    return;
}

/* unblock signals temporarily so we can receive pending signals. */
int check_for_interrupts(void)
{
    /* Easiest way to do this is unblock all signals,
     * then block them again. */
    
    sigset_t signal_mask, o_mask;
    
    sigemptyset(&signal_mask);
    
    sigprocmask(SIG_SETMASK, &signal_mask, &o_mask);
    /* here we get all the signals. */
    sigprocmask(SIG_SETMASK, &o_mask, NULL);
    
    return 0;
}

/* suspend indefinetly for a signal (useful for alarm). */
void suspend_for_interrupts(void)
{
    sigset_t signal_mask;
    
    sigemptyset(&signal_mask);
    sigsuspend(&signal_mask);
    
}

void remove_interrupt(int sig)
{
    sigset_t signal_set;

    sigemptyset(&signal_set);
    sigaddset(&signal_set, sig);

    if(sigpending(&signal_set)) {
        add_interrupt_handler(sig, SIG_IGN);
        sigfillset(&signal_set);
        sigdelset(&signal_set, SIGALRM);

        sigsuspend(&signal_set);
        add_interrupt_handler(sig, SIG_DFL);
    }

    return;
}

/* String routines. */

/* 
 * Splice two strings together.
 * 
 */
    
char *splice_string(const char *s1, const char *s2)
{
    char *new_string;

    if(s1 == NULL)
        return(strdup(s2));

    if(s2 == NULL)
        return(strdup(s1));
    
    new_string = xmalloc(strlen(s1) + strlen(s2) + 1);
    strcpy(new_string, s1);
    strcat(new_string, s2);

    return new_string;
}

/* splice many strings lets us pass a variable amount of strings to be spliced. */
char *splice_many_strings(int num, char *string, ...)
{
    va_list ap;
    char *s, *tmp;
    char *new_string;
    
    new_string = strdup(string);
    
    if(num <= 1)
        return new_string; /* be nice to stupid callers. */

    va_start(ap, string);
    
    while(num--)  {
        s = va_arg(ap, char *);
        tmp = new_string;
        new_string = splice_string(new_string, s);
        xfree(tmp);
    }
    
    va_end(ap);

    return new_string;
}
             
/* trim preprending and trailing space. also get rid of newlines. */
void trim_string(char *s)
{
    char *start, *end, *tmp_string;
    int len;
    
    len = strlen(s);
    
    if(len == 0)
        return;
    
    tmp_string = xmalloc(len + 1);
    strcpy(tmp_string, s);
    tmp_string[len] = '\0';

    start = tmp_string;

    while(isspace(*start))
        start++;

    end = &tmp_string[len - 1];
    
    while((end == '\0' || isspace(*end)) && end != start)  {
        end--;
    }
    
    if(end == start) /* bad. */
        s[0] = '\0';

    end++;
    *end = '\0';
    
    strcpy(s, start);
    xfree(tmp_string);
    
    return;
}

/* is a valid C string? */
int is_string(const char *string, int len)
{
    const char *cp;
    int i = 0;
    
    cp = string;
    while(i < len)  {
        if(*cp == '\0')
            return 1;
        i++;
        cp++;
    }
    
    return 0;
}

/* parse hex string (0x:0x...0x) to value. (not fun to write) */
int hex_string_to_value(char *string, unsigned char *dst)
{
    char *sep, *ptr;
    char *new_string;
    char *val_string;
    int len, i;
    
    len = strlen(string);
    if(len < 17)
        return -1;
    if(len == 17)
        new_string = splice_string(string, ":");
    else
        new_string = strdup(string);

    if(strlen(new_string) > 18)  {
        xfree(new_string);

        return -1;
    }
    
    i = 0;
    ptr = new_string;
    while((sep = strchr(ptr, ':')) != NULL)  {

        *sep = 0;
        val_string = splice_string(" ", ptr);
        dst[i] = strtoul(val_string, NULL, 16);
        xfree(val_string);
        i++;
        ptr = sep + 1;
        
    }
    
    xfree(new_string);
    return 0;
}

/* check if two strings match. */
int string_matches(const char *s1, const char *s2)
{
    if(!strcmp(s1, s2))
        return 1;
    else
        return 0;
}

static rand_t *ran_gen = NULL;

/* initialize random seed. */
static void init_rand(void)
{
    if(ran_gen == NULL)
        if((ran_gen = rand_open()) == NULL)
            fatal_error("unable to access random number generator");

    return;
}

uint16_t get_random_uint16(void)
{
    init_rand(); /* reinit if needed. */
    return(rand_uint16(ran_gen));
}

uint32_t get_random_uint32(void)
{
    init_rand(); /* reinit if needed. */
    return(rand_uint32(ran_gen));
}

struct timeval timeval_diff(struct timeval begin, struct timeval end)
{
    struct timeval difference;

    if(begin.tv_usec > end.tv_usec) {
        end.tv_sec--;
        end.tv_usec += 1000000;
    }
  
    difference.tv_sec = end.tv_sec - begin.tv_sec;
    difference.tv_usec = end.tv_usec - begin.tv_usec;
  
    return difference;
}

#if !defined(HAVE_GETPROGNAME)
const char *getprogname(void)
{
    /* 
     * Compatibility looks like this:
     * if getprogname this will not be.
     * if __progname is available use it here.
     * if __progname is not available set it in main source,
     *               and use it here.
     */

    /* FIXME: do the system provided getprogname() do this? */
    if(__progname[0] == '.' && __progname[1] == '/')
        return (__progname + 2);
    else 
        return(__progname);

}

#endif /* HAVE_GETPROGNAME */

/* alarm handling routine and variable. */

/* wrap for volatine integers, or proper sig_atomic_t types. */

#ifdef HAVE_SIG_ATOMIC_T

static sig_atomic_t have_alarm = 0;

#else /* HAVE_SIG_ATOMIC_T */

static volatile int have_alarm = 0;

#endif

static RETSIGTYPE handle_alarm(int i)
{
    have_alarm = 1;
    return (RETSIGTYPE) 0;
}

static struct timeval disarm_timer  = { 0, 0 };

/* will trigger alarm for one time only. */
void set_alarm(struct timeval alarm_time)
{
    struct itimerval timer;

    timer.it_interval = alarm_time;
    timer.it_value = disarm_timer;

    have_alarm = 0;
    add_interrupt_handler(SIGALRM, handle_alarm);

    /* that's it. after this the timer is ticking. */
    setitimer(ITIMER_REAL, &timer, NULL);

    return;
}

/* tells us whether we received an alarm or not.*/
int had_alarm(void)
{
    return have_alarm;
}

void remove_alarm(void)
{
    struct itimerval timer;

    timer.it_interval = disarm_timer;
    timer.it_value = disarm_timer;

    setitimer(ITIMER_REAL, &timer, NULL);

    /* now make sure we don't have a pending SIGARLM. */
    remove_interrupt(SIGALRM);

    return;
}

/* sigio handling routine and variable.
 * we only need to use this on *BSD
 * with broken BPF which can't handle a select(). */

/* wrap for volatine integers, or proper sig_atomic_t types. */

#ifdef HAVE_SIG_ATOMIC_T

static sig_atomic_t have_io = 0;

#else /* HAVE_SIG_ATOMIC_T */

static volatile int have_io = 0;

#endif

static RETSIGTYPE handle_sigio(int i)
{
    have_io = 1;
}

void track_sigio(void)
{
    have_io = 0;
    add_interrupt_handler(SIGIO, handle_sigio);
    return;
}

int had_io(void)
{
    return have_io;
}

/* Get port number for named service. */

int port_for_service(const char *serv, const char *proto)
{
    struct servent *sv;

    setservent(0);

    if((sv = getservbyname(serv, proto)) == NULL)  {
        error_message("rawnet: could not get service listing: %d", strerror(errno));
        return -1;
    }
    
    return(sv->s_port);
}

/* Returns 0 on success or -1 if error.
 * Fills in addr with 32-bit ipv4 address */

extern int h_errno;

int resolv(char *address, uint32_t *addr)
{
    struct hostent *host;

    *(addr) = inet_addr(address);

    if(*(addr) == INADDR_NONE)  {
        
        host = gethostbyname(address);

        if(host == NULL)  {
            error_message("rawnet: unable to resolv host: %s", hstrerror(h_errno));
            return -1;
        }
        
        memcpy(&addr, (int *) * &host->h_addr, sizeof(addr));
    }

    return 0;
}
